# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import, g-importing-member
"""Base brax primitives and simple manipulations of them."""

import copy
import functools
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from brax import math
from flax import struct
import jax
from jax import numpy as jp
from jax import vmap
from jax.tree_util import tree_map
import mujoco
from mujoco import mjx
import numpy as np

# f: free, 1: 1-dof, 2: 2-dof, 3: 3-dof
Q_WIDTHS = {'f': 7, '1': 1, '2': 2, '3': 3}
QD_WIDTHS = {'f': 6, '1': 1, '2': 2, '3': 3}


@struct.dataclass
class Base:
  """Base functionality extending all brax types.

  These methods allow for brax types to be operated like arrays/matrices.
  """

  def __add__(self, o: Any) -> Any:
    return tree_map(lambda x, y: x + y, self, o)

  def __sub__(self, o: Any) -> Any:
    return tree_map(lambda x, y: x - y, self, o)

  def __mul__(self, o: Any) -> Any:
    return tree_map(lambda x: x * o, self)

  def __neg__(self) -> Any:
    return tree_map(lambda x: -x, self)

  def __truediv__(self, o: Any) -> Any:
    return tree_map(lambda x: x / o, self)

  def reshape(self, shape: Sequence[int]) -> Any:
    return tree_map(lambda x: x.reshape(shape), self)

  def select(self, o: Any, cond: jax.Array) -> Any:
    return tree_map(lambda x, y: (x.T * cond + y.T * (1 - cond)).T, self, o)

  def slice(self, beg: int, end: int) -> Any:
    return tree_map(lambda x: x[beg:end], self)

  def take(self, i, axis=0) -> Any:
    return tree_map(lambda x: jp.take(x, i, axis=axis, mode='wrap'), self)

  def concatenate(self, *others: Any, axis: int = 0) -> Any:
    return tree_map(lambda *x: jp.concatenate(x, axis=axis), self, *others)

  def index_set(
      self, idx: Union[jax.Array, Sequence[jax.Array]], o: Any
  ) -> Any:
    return tree_map(lambda x, y: x.at[idx].set(y), self, o)

  def index_sum(
      self, idx: Union[jax.Array, Sequence[jax.Array]], o: Any
  ) -> Any:
    return tree_map(lambda x, y: x.at[idx].add(y), self, o)

  def vmap(self, in_axes=0, out_axes=0):
    """Returns an object that vmaps each follow-on instance method call."""

    # TODO: i think this is kinda handy, but maybe too clever?

    outer_self = self

    class VmapField:
      """Returns instance method calls as vmapped."""

      def __init__(self, in_axes, out_axes):
        self.in_axes = [in_axes]
        self.out_axes = [out_axes]

      def vmap(self, in_axes=0, out_axes=0):
        self.in_axes.append(in_axes)
        self.out_axes.append(out_axes)
        return self

      def __getattr__(self, attr):
        fun = getattr(outer_self.__class__, attr)
        # load the stack from the bottom up
        vmap_order = reversed(list(zip(self.in_axes, self.out_axes)))
        for in_axes, out_axes in vmap_order:
          fun = vmap(fun, in_axes=in_axes, out_axes=out_axes)
        fun = functools.partial(fun, outer_self)
        return fun

    return VmapField(in_axes, out_axes)

  def tree_replace(
      self, params: Dict[str, Optional[jax.typing.ArrayLike]]
  ) -> 'Base':
    """Creates a new object with parameters set.

    Args:
      params: a dictionary of key value pairs to replace

    Returns:
      data clas with new values

    Example:
      If a system has 3 links, the following code replaces the mass
      of each link in the System:
      >>> sys = sys.tree_replace(
      >>>     {'link.inertia.mass', jp.array([1.0, 1.2, 1.3])})
    """
    new = self
    for k, v in params.items():
      new = _tree_replace(new, k.split('.'), v)
    return new

  @property
  def T(self):  # pylint:disable=invalid-name
    return tree_map(lambda x: x.T, self)


def _tree_replace(
    base: Base,
    attr: Sequence[str],
    val: Optional[jax.typing.ArrayLike],
) -> Base:
  """Sets attributes in a struct.dataclass with values."""
  if not attr:
    return base

  # special case for List attribute
  if len(attr) > 1 and isinstance(getattr(base, attr[0]), list):
    lst = copy.deepcopy(getattr(base, attr[0]))

    for i, g in enumerate(lst):
      if not hasattr(g, attr[1]):
        continue
      v = val if not hasattr(val, '__iter__') else val[i]
      lst[i] = _tree_replace(g, attr[1:], v)

    return base.replace(**{attr[0]: lst})

  if len(attr) == 1:
    return base.replace(**{attr[0]: val})

  return base.replace(
      **{attr[0]: _tree_replace(getattr(base, attr[0]), attr[1:], val)}
  )


@struct.dataclass
class Transform(Base):
  """Transforms the position and rotation of a coordinate frame.

  Attributes:
    pos: (3,) position transform of the coordinate frame
    rot: (4,) quaternion rotation the coordinate frame
  """

  pos: jax.Array
  rot: jax.Array

  def do(self, o):
    """Apply the transform."""
    return _transform_do(o, self)

  def inv_do(self, o):
    """Apply the inverse of the transform."""
    return _transform_inv_do(o, self)

  def to_local(self, t: 'Transform') -> 'Transform':
    """Move transform into basis of t."""
    pos = math.rotate(self.pos - t.pos, math.quat_inv(t.rot))
    rot = math.quat_mul(math.quat_inv(t.rot), self.rot)
    return Transform(pos=pos, rot=rot)

  @classmethod
  def create(
      cls, pos: Optional[jax.Array] = None, rot: Optional[jax.Array] = None
  ) -> 'Transform':
    """Creates a transform with either pos, rot, or both."""
    if pos is None and rot is None:
      raise ValueError('must specify either pos or rot')
    elif pos is None and rot is not None:
      pos = jp.zeros(rot.shape[:-1] + (3,))
    elif rot is None and pos is not None:
      rot = jp.tile(jp.array([1.0, 0.0, 0.0, 0.0]), pos.shape[:-1] + (1,))
    return Transform(pos=pos, rot=rot)

  @classmethod
  def zero(cls, shape=()) -> 'Transform':
    """Returns a zero transform with a batch shape."""
    pos = jp.zeros(shape + (3,))
    rot = jp.tile(jp.array([1.0, 0.0, 0.0, 0.0]), shape + (1,))
    return Transform(pos, rot)


@struct.dataclass
class Motion(Base):
  """Spatial motion vector describing linear and angular velocity.

  More on spatial vectors: http://royfeatherstone.org/spatial/v2/index.html

  Attributes:
    ang: (3,) angular velocity about a normal
    vel: (3,) linear velocity in the direction of the normal
  """

  ang: jax.Array
  vel: jax.Array

  def cross(self, other):
    return _motion_cross(other, self)

  def dot(self, m: Union['Motion', 'Force']) -> jax.Array:
    return jp.dot(self.vel, m.vel) + jp.dot(self.ang, m.ang)

  def matrix(self) -> jax.Array:
    return jp.concatenate([self.ang, self.vel], axis=-1)

  @classmethod
  def create(
      cls, ang: Optional[jax.Array] = None, vel: Optional[jax.Array] = None
  ) -> 'Motion':
    if ang is None and vel is None:
      raise ValueError('must specify either ang or vel')
    ang = jp.zeros_like(vel) if ang is None else ang
    vel = jp.zeros_like(ang) if vel is None else vel

    return Motion(ang=ang, vel=vel)

  @classmethod
  def zero(cls, shape=()) -> 'Motion':
    ang = jp.zeros(shape + (3,))
    vel = jp.zeros(shape + (3,))
    return Motion(ang, vel)


@struct.dataclass
class Force(Base):
  """Spatial force vector describing linear and angular (torque) force.

  Attributes:
    ang: (3,) angular velocity about a normal
    vel: (3,) linear velocity in the direction of the normal
  """

  ang: jax.Array
  vel: jax.Array

  @classmethod
  def create(
      cls, ang: Optional[jax.Array] = None, vel: Optional[jax.Array] = None
  ) -> 'Force':
    if ang is None and vel is None:
      raise ValueError('must specify either ang or vel')
    ang = jp.zeros_like(vel) if ang is None else ang
    vel = jp.zeros_like(ang) if vel is None else vel

    return Force(ang=ang, vel=vel)


@struct.dataclass
class Inertia(Base):
  """Angular inertia, mass, and center of mass location.

  Attributes:
    transform: transform for the inertial frame relative to the link frame
      (i.e. center of mass position and orientation)
    i: (3, 3) inertia matrix about a point P
    mass: scalar mass
  """

  transform: Transform
  i: jax.Array
  mass: jax.Array

  def mul(self, m: Motion) -> 'Force':
    """Multiplies inertia with motion yielding a force."""
    ang = jp.dot(self.i, m.ang) + jp.cross(self.transform.pos, m.vel)
    vel = self.mass * m.vel - jp.cross(self.transform.pos, m.ang)

    return Force(ang=ang, vel=vel)


@struct.dataclass
class Link(Base):
  """A rigid segment of an articulated body.

  Links are connected to each other by joints.  By moving (rotating or
  translating) the joints, the entire system can be articulated.

  Attributes:
    transform: transform for the link frame relative to the parent frame
    joint: transform for the joint frame relative to the link frame
    inertia: mass, center of mass location, and inertia of this link
    invweight: mean inverse inertia at init_q
    constraint_stiffness: (num_link,) constraint spring for joint.
    constraint_vel_damping: (num_link,) linear damping for constraint spring.
    constraint_limit_stiffness: (num_link,) constraint for angle limits
    constraint_ang_damping: (num_link,) angular damping for constraint spring.
  """

  transform: Transform
  joint: Transform
  inertia: Inertia
  invweight: jax.Array
  # only used by `brax.physics.spring`:
  constraint_stiffness: jax.Array
  constraint_vel_damping: jax.Array
  constraint_limit_stiffness: jax.Array
  # only used by `brax.physics.spring` and `brax.physics.positional`:
  constraint_ang_damping: jax.Array


@struct.dataclass
class DoF(Base):
  """A degree of freedom in the system.

  Attributes:
    motion: spatial motion (linear or angular) of this DoF
    armature: models the inertia of a rotor (moving part of a motor)
    stiffness: restorative force back to zero position
    damping: restorative force back to zero velocity
    limit: tuple of min, max angle limits
    invweight: diagonal inverse inertia at init_qpos
    solver_params: (7,) limit constraint solver parameters
  """

  motion: Motion
  armature: jax.Array
  stiffness: jax.Array
  damping: jax.Array
  limit: Tuple[jax.Array, jax.Array]
  # only used by `brax.physics.generalized`:
  invweight: jax.Array
  solver_params: jax.Array


class Contact(mjx.Contact, Base):
  """Contact between two geometries.

  Attributes:
    link_idx: Tuple of link indices participating in contact.
    elasticity: bounce/restitution encountered when hitting another geometry
  """

  link_idx: jax.Array
  elasticity: jax.Array


@struct.dataclass
class Actuator(Base):
  """Actuator, transforms an input signal into a force (motor or thruster).

  Attributes:
    q_id: (num_actuators,) q index associated with an actuator
    qd_id: (num_actuators,) qd index associated with an actuator
    ctrl_range: (num_actuators, 2) actuator control range
    force_range: (num_actuators, 2) actuator force range
    gain: (num_actuators,) scaling factor for each actuator control input
    gear: (num_actuators,) scaling factor for each actuator force output
    bias_q: (num_actuators,) bias applied by q (e.g. position actuators)
    bias_qd: (num_actuators,) bias applied by qd (e.g. velocity actuators)
  """

  q_id: jax.Array
  qd_id: jax.Array
  ctrl_range: jax.Array
  force_range: jax.Array
  gain: jax.Array
  gear: jax.Array
  bias_q: jax.Array
  bias_qd: jax.Array


@struct.dataclass
class State:
  """Dynamic state that changes after every pipeline step.

  Attributes:
    q: (q_size,) joint position vector
    qd: (qd_size,) joint velocity vector
    x: (num_links,) link position in world frame
    xd: (num_links,) link velocity in world frame
    contact: calculated contacts
  """

  q: jax.Array
  qd: jax.Array
  x: Transform
  xd: Motion
  contact: Optional[Contact]


class System(mjx.Model):
  r"""Describes a physical environment: its links, joints and geometries.

  Attributes:
    gravity: (3,) linear universal force applied during forward dynamics
    viscosity: (1,) viscosity of the medium applied to all links
    density: (1,) density of the medium applied to all links
    link: (num_link,) the links in the system
    dof: (qd_size,) every degree of freedom for the system
    actuator: actuators that can be applied to links
    init_q: (q_size,) initial q position for the system
    elasticity: bounce/restitution encountered when hitting another geometry
    vel_damping: (1,) linear vel damping applied to each body.
    ang_damping: (1,) angular vel damping applied to each body.
    baumgarte_erp: how aggressively interpenetrating bodies should push away\
                from one another
    spring_mass_scale: a float that scales mass as `mass^(1 - x)`
    spring_inertia_scale: a float that scales inertia diag as `inertia^(1 - x)`
    joint_scale_ang: scale for position-based joint rotation update
    joint_scale_pos: scale for position-based joint position update
    collide_scale: fraction of position based collide update to apply
    enable_fluid: (1,) enables or disables fluid forces based on the
      default viscosity and density parameters provided in the XML
    link_names: (num_link,) link names
    link_types: (num_link,) string specifying the joint type of each link
                valid types are:
                * 'f': free, full 6 dof (position + rotation), no parent link
                * '1': revolute,  1 dof, like a hinge
                * '2': universal, 2 dof, like a drive shaft joint
                * '3': spherical, 3 dof, like a ball joint
    link_parents: (num_link,) int list specifying the index of each link's
                  parent link, or -1 if the link has no parent
    matrix_inv_iterations: maximum number of iterations of the matrix inverse
    solver_iterations: maximum number of iterations of the constraint solver
    solver_maxls: maximum number of line searches of the constraint solver
    mj_model: mujoco.MjModel that was used to build this brax System
  """

  gravity: jax.Array
  viscosity: Union[float, jax.Array]
  density: Union[float, jax.Array]
  link: Link
  dof: DoF
  actuator: Actuator
  init_q: jax.Array
  # only used in `brax.physics.spring` and `brax.physics.positional`:
  elasticity: jax.Array
  vel_damping: Union[float, jax.Array]
  ang_damping: Union[float, jax.Array]
  baumgarte_erp: Union[float, jax.Array]
  spring_mass_scale: Union[float, jax.Array]
  spring_inertia_scale: Union[float, jax.Array]
  # only used in `brax.physics.positional`:
  joint_scale_ang: Union[float, jax.Array]
  joint_scale_pos: Union[float, jax.Array]
  collide_scale: Union[float, jax.Array]
  # non-pytree nodes
  enable_fluid: bool = struct.field(pytree_node=False)
  link_names: List[str] = struct.field(pytree_node=False)
  link_types: str = struct.field(pytree_node=False)
  link_parents: Tuple[int, ...] = struct.field(pytree_node=False)
  # only used in `brax.physics.generalized`:
  matrix_inv_iterations: int = struct.field(pytree_node=False)
  solver_iterations: int = struct.field(pytree_node=False)
  solver_maxls: int = struct.field(pytree_node=False)
  mj_model: mujoco.MjModel = struct.field(pytree_node=False, default=None)

  def num_links(self) -> int:
    """Returns the number of links in the system."""
    return len(self.link_types)

  def dof_link(self, depth=False) -> jax.Array:
    """Returns the link index corresponding to each system dof."""
    link_idxs = []
    for i, link_type in enumerate(self.link_types):
      link_idxs.extend([i] * QD_WIDTHS[link_type])
    if depth:
      depth_fn = lambda i, p=self.link_parents: p[i] + 1 and 1 + depth_fn(p[i])
      depth_count = {}
      link_idx_depth = []
      for i in range(self.num_links()):
        depth = depth_fn(i)
        depth_idx = depth_count.get(depth, 0)
        depth_count[depth] = depth_idx + 1
        link_idx_depth.append(depth_idx)
      link_idxs = [link_idx_depth[i] for i in link_idxs]

    return jp.array(link_idxs)

  def dof_ranges(self) -> List[List[int]]:
    """Returns the dof ranges corresponding to each link."""
    beg, ranges = 0, []
    for t in self.link_types:
      ranges.append(list(range(beg, beg + QD_WIDTHS[t])))
      beg += QD_WIDTHS[t]
    return ranges

  def q_idx(self, link_type: str) -> jax.Array:
    """Returns the q indices corresponding to a link type."""
    idx, idxs = 0, []
    for typ in self.link_types:
      if typ in link_type:
        idxs.extend(range(idx, idx + Q_WIDTHS[typ]))
      idx += Q_WIDTHS[typ]
    return jp.array(idxs)

  def qd_idx(self, link_type: str) -> jax.Array:
    """Returns the qd indices corresponding to a link type."""
    idx, idxs = 0, []
    for typ in self.link_types:
      if typ in link_type:
        idxs.extend(range(idx, idx + QD_WIDTHS[typ]))
      idx += QD_WIDTHS[typ]
    return jp.array(idxs)

  def q_size(self) -> int:
    """Returns the size of the q vector (joint position) for this system."""
    return self.nq

  def qd_size(self) -> int:
    """Returns the size of the qd vector (joint velocity) for this system."""
    return self.nv

  def act_size(self) -> int:
    """Returns the act dimension for the system."""
    return self.nu


# below are some operation dispatch derivations


@functools.singledispatch
def _transform_do(other, self: Transform):
  del other, self
  return NotImplemented


@functools.singledispatch
def _transform_inv_do(other, self: Transform):
  del other, self
  return NotImplemented


@_transform_do.register(Transform)
def _(t: Transform, self: Transform) -> Transform:
  pos = self.pos + math.rotate(t.pos, self.rot)
  rot = math.quat_mul(self.rot, t.rot)
  return Transform(pos, rot)


@_transform_do.register(Motion)
def _(m: Motion, self: Transform) -> Motion:
  rot_t = math.quat_inv(self.rot)
  ang = math.rotate(m.ang, rot_t)
  vel = math.rotate(m.vel - jp.cross(self.pos, m.ang), rot_t)
  return Motion(ang, vel)


@_transform_inv_do.register(Motion)
def _(m: Motion, self: Transform) -> Motion:
  rot_t = self.rot
  ang = math.rotate(m.ang, rot_t)
  vel = math.rotate(m.vel, rot_t) + jp.cross(self.pos, ang)
  return Motion(ang, vel)


@_transform_do.register(Force)
def _(f: Force, self: Transform) -> Force:
  vel = math.rotate(f.vel, self.rot)
  ang = math.rotate(f.ang, self.rot) + jp.cross(self.pos, vel)
  return Force(ang, vel)


@_transform_do.register(Inertia)
def _(it: Inertia, self: Transform) -> Inertia:
  h = jp.cross(self.pos, -jp.eye(3))
  rot = math.quat_to_3x3(self.rot)
  i = rot @ it.i @ rot.T + h @ h.T * it.mass
  transform = Transform(pos=self.pos * it.mass, rot=self.rot)
  return Inertia(transform=transform, i=i, mass=it.mass)


@functools.singledispatch
def _motion_cross(other, self: Motion):
  del other, self
  return NotImplemented


@_motion_cross.register(Motion)
def _(m: Motion, self: Motion) -> Motion:
  vel = jp.cross(self.ang, m.vel) + jp.cross(self.vel, m.ang)
  ang = jp.cross(self.ang, m.ang)
  return Motion(ang, vel)


@_motion_cross.register(Force)
def _(f: Force, self: Motion) -> Force:
  vel = jp.cross(self.ang, f.vel)
  ang = jp.cross(self.ang, f.ang) + jp.cross(self.vel, f.vel)
  return Force(ang, vel)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of BIG-Gym environments and components."""

# keep alphabetical ordering
REGISTRIES = [
    'jump',
    'proant',
    'ant',
]


def get_comp_name(registry_name: str, comp_name: str):
  return f'{registry_name}__{comp_name}'


def get_env_name(registry_name: str, env_name: str):
  return f'{registry_name}__{env_name}'



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Replay stored trajectories and simulate brax systems."""

from wsgiref import simple_server
from wsgiref import validate

from absl import app
from absl import flags
from brax.base import Actuator
from brax.generalized import pipeline as generalized
from brax.io import html
from brax.io import mjcf
from brax.io.mjcf import load_mjmodel
from brax.positional import pipeline as positional
from brax.spring import pipeline as spring
from etils import epath
import flask
from flask import jsonify
from flask import request
from flask import send_from_directory
import flask_cors
import jax
from jax import numpy as jp
import mujoco


PORT = flags.DEFINE_integer(
    name='port', default=8080, help='Port to run server on'
)
DEBUG = flags.DEFINE_boolean(
    name='debug', default=False, help='Debug the server.'
)

flask_app = flask.Flask(__name__)
flask_cors.CORS(flask_app)


class _MujocoPipeline:
  """Mujoco reference pipeline for debugging."""

  @staticmethod
  def init(*args, **kwargs):
    return generalized.init(*args, **kwargs)

  @staticmethod
  def step(*args, **kwargs):
    return generalized.init(*args, **kwargs)


@flask_app.route('/', methods=['GET'])
def index():
  return jsonify(success=True)


@flask_app.route('/favicon.ico')
def favicon():
  """Serves the brax favicon."""
  path = epath.Path(flask_app.root_path)
  return send_from_directory(str(path), 'favicon.ico')


@flask_app.route('/js/<path:path>', methods=['GET'])
def js(path):
  """Serves files from the js/ directory."""
  path = epath.Path(flask_app.root_path) / 'js' / path
  response = flask.Response(path.read_text(), mimetype='text/javascript')
  response.headers.add('Access-Control-Allow-Origin', '*')
  return response


@flask_app.route('/play/<path:path>', methods=['GET'])
def play_trajectory(path):
  """Renders a json-encoded brax trajectory from a local file path."""
  system = epath.Path(path).read_text()
  return html.render_from_json(
      system, height='100vh', colab=False, base_url='/js/viewer.js'
  )


@flask_app.route('/simulate/<path:path>', methods=['GET'])
def simulate(path):
  """Simulates a brax system from a local file path."""
  sys = mjcf.load(path)
  pipeline_type = request.args.get('pipeline', 'generalized')
  pipeline = {
      'generalized': generalized,
      'positional': positional,
      'spring': spring,
      'mujoco': _MujocoPipeline,
  }[pipeline_type]
  steps = int(request.args.get('steps', 1000))
  act_fn = request.args.get('act', 'sin')
  solver_iterations = int(request.args.get('solver_iterations', 100))
  add_act = request.args.get('add_act', 'false').lower() == 'true'

  if solver_iterations > 0:
    sys = sys.replace(solver_iterations=solver_iterations)
  if add_act and not sys.actuator_types:
    # some configs (like urdfs) have no actuators
    actuator_types = ''.join(['m' for t in sys.link_types if t in '123'])
    actuator_link_id = [i for i, t in enumerate(sys.link_types) if t in '123']
    actuator_qid = [int(i) for i in sys.q_idx('123')]
    actuator_qdid = [int(i) for i in sys.qd_idx('123')]
    actuator = Actuator(
        ctrl_range=jp.tile(jp.array([-1.0, 1.0]), (len(actuator_types), 2)),
        gear=25 * jp.ones(len(actuator_types)),
    )
    sys = sys.replace(
        actuator=actuator,
        actuator_types=actuator_types,
        actuator_link_id=actuator_link_id,
        actuator_qid=actuator_qid,
        actuator_qdid=actuator_qdid,
    )

  if pipeline_type == 'mujoco':
    mj_model = load_mjmodel(path)
    mj_data = mujoco.MjData(mj_model)
    init_fn = jax.jit(pipeline.init)
    def step_fn(sys, _, act):
      mj_data.ctrl = act
      mujoco.mj_step(mj_model, mj_data)
      state = init_fn(sys, mj_data.qpos, jp.zeros(sys.qd_size()))
      return state
  else:
    init_fn, step_fn = jax.jit(pipeline.init), jax.jit(pipeline.step)
  state = init_fn(sys, sys.init_q, jp.zeros(sys.qd_size()))
  states = [state]

  for i in range(steps):
    if act_fn == 'sin':
      act = 0.5 * jp.sin(jp.ones(sys.act_size()) * 5 * i * sys.opt.timestep)
    elif act_fn == 'zero':
      act = jp.zeros(sys.act_size())
    elif act_fn == 'zero_p':
      q = states[-1].q[sys.q_idx('123')]
      act = -q
    else:
      raise ValueError(f'Unknown act function: {act_fn}')
    state = step_fn(sys, states[-1], act)
    states.append(state)
  return html.render(
      sys, states, height='100vh', colab=False, base_url='/js/viewer.js'
  )


def main(_):
  if DEBUG.value:
    flask_app.run(
        host='localhost', port=PORT.value, use_reloader=True, debug=True
    )
    return

  server = simple_server.make_server(
      'localhost', PORT.value, validate.validator(flask_app)
  )
  server.serve_forever()


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Integrator functions."""

from brax import math
from brax import scan
from brax.base import System
from brax.generalized.base import State
import jax
from jax import numpy as jp


def _integrate_q_axis(sys: System, q: jax.Array, qd: jax.Array) -> jax.Array:
  """Integrates next q for revolute/prismatic joints."""
  return q + qd * sys.opt.timestep


def _integrate_q_free(sys: System, q: jax.Array, qd: jax.Array) -> jax.Array:
  """Integrates next q for free joints."""
  rot, ang = q[3:7], qd[3:6]
  ang_norm = jp.linalg.norm(ang) + 1e-8
  axis = ang / ang_norm
  angle = sys.opt.timestep * ang_norm
  qrot = math.quat_rot_axis(axis, angle)
  rot = math.quat_mul(rot, qrot)
  rot = rot / jp.linalg.norm(rot)
  pos, vel = q[0:3], qd[0:3]
  pos += vel * sys.opt.timestep

  return jp.concatenate([pos, rot])


def integrate(sys: System, state: State) -> State:
  """Semi-implicit Euler integration.

  Args:
    sys: system defining the kinematic tree and other properties
    state: generalized state

  Returns:
    state: state with q, qd, and qdd updated
  """
  # integrate joint damping implicitly to increase stability when we are not
  # using approximate inverse
  if sys.matrix_inv_iterations == 0:
    mx = state.mass_mx + jp.diag(sys.dof.damping) * sys.opt.timestep
    mx_inv = jax.scipy.linalg.solve(mx, jp.eye(sys.qd_size()), assume_a='pos')
  else:
    mx_inv = state.mass_mx_inv
  qdd = mx_inv @ (state.qf_smooth + state.qf_constraint)
  qd = state.qd + qdd * sys.opt.timestep

  def q_fn(typ, link, q, qd):
    q = q.reshape(link.transform.pos.shape[0], -1)
    qd = qd.reshape(link.transform.pos.shape[0], -1)
    fun = jax.vmap(
        {
            'f': _integrate_q_free,
            '1': _integrate_q_axis,
            '2': _integrate_q_axis,
            '3': _integrate_q_axis,
        }[typ],
        in_axes=(None, 0, 0),
    )
    q_s = fun(sys, q, qd).reshape(-1)

    return q_s

  q = scan.link_types(sys, q_fn, 'lqd', 'q', sys.link, state.q, qd)

  return state.replace(q=q, qd=qd, qdd=qdd)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Gotta go fast!  This trivial Env is for unit testing."""

from brax import base
from brax.envs.base import PipelineEnv, State
import jax
from jax import numpy as jp


class Fast(PipelineEnv):
  """Trains an agent to go fast."""

  def __init__(self, **kwargs):
    self._dt = 0.02
    self._reset_count = 0
    self._step_count = 0

  def reset(self, rng: jax.Array) -> State:
    self._reset_count += 1
    pipeline_state = base.State(
        q=jp.zeros(1),
        qd=jp.zeros(1),
        x=base.Transform.create(pos=jp.zeros(3)),
        xd=base.Motion.create(vel=jp.zeros(3)),
        contact=None
    )
    obs = jp.zeros(2)
    reward, done = jp.array(0.0), jp.array(0.0)
    return State(pipeline_state, obs, reward, done)

  def step(self, state: State, action: jax.Array) -> State:
    assert state.pipeline_state is not None
    self._step_count += 1
    vel = state.pipeline_state.xd.vel + (action > 0) * self._dt
    pos = state.pipeline_state.x.pos + vel * self._dt

    qp = state.pipeline_state.replace(
        x=state.pipeline_state.x.replace(pos=pos),
        xd=state.pipeline_state.xd.replace(vel=vel),
    )
    obs = jp.array([pos[0], vel[0]])
    reward = pos[0]

    return state.replace(pipeline_state=qp, obs=obs, reward=reward)

  @property
  def reset_count(self):
    return self._reset_count

  @property
  def step_count(self):
    return self._step_count

  @property
  def observation_size(self):
    return 2

  @property
  def action_size(self):
    return 1



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for mass matrices."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import test_utils
from brax.generalized import pipeline
import jax
import mujoco
import numpy as np


class MassTest(parameterized.TestCase):

  @parameterized.parameters(
      ('ant.xml',),
      ('triple_pendulum.xml',),
      ('humanoid.xml',),
      ('half_cheetah.xml',),
  )
  def test_matrix(self, xml_file):
    """Test mass matrix calculation."""
    sys = test_utils.load_fixture(xml_file)
    model = test_utils.load_fixture_mujoco(xml_file)
    mj_mass_mx = np.zeros((sys.qd_size(), sys.qd_size()))

    for mj_prev, mj_next in test_utils.sample_mujoco_states(xml_file):
      state = jax.jit(pipeline.init)(sys, mj_prev.qpos, mj_prev.qvel)
      mujoco.mj_fullM(model, mj_mass_mx, mj_next.qM)
      np.testing.assert_almost_equal(state.mass_mx, mj_mass_mx, 5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multi-agent proximal policy optimization training.

*This is branched from brax.v1lines/training/ppo.py, and will be folded back.*
"""

from collections import OrderedDict as odict
import functools
import time
from typing import Any, Callable, Dict, Optional, Tuple

from absl import logging
from brax.v1 import envs
from brax.v1.experimental import normalization
from brax.v1.experimental.braxlines.training import env
from brax.v1.experimental.composer import data_utils
from brax.training import distribution
from brax.training import networks
from brax.training import pmap
from brax.training.agents.ppo import losses as ppo_losses
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
import jax
import jax.numpy as jnp
import optax


@flax.struct.dataclass
class StepData:
  """Contains data for one environment step."""
  obs: jnp.ndarray
  rewards: jnp.ndarray
  dones: jnp.ndarray
  truncation: jnp.ndarray
  actions: jnp.ndarray
  logits: jnp.ndarray


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  optimizer_state: optax.OptState
  params: Params
  key: PRNGKey
  normalizer_params: Params


@flax.struct.dataclass
class Agent:
  parametric_action_distribution: distribution.ParametricDistribution
  policy_model: Any
  optimizer_state: Any
  init_params: Any
  grad_loss: Any


def compute_ppo_loss(
    models: Dict[str, Params],
    data: StepData,
    udata: StepData,
    rng: PRNGKey,
    parametric_action_distribution: distribution.ParametricDistribution,
    policy_apply: Any,
    value_apply: Any,
    entropy_cost: float = 1e-4,
    discounting: float = 0.9,
    reward_scaling: float = 1.0,
    lambda_: float = 0.95,
    ppo_epsilon: float = 0.3,
    extra_loss_update_ratios: Optional[Dict[str, float]] = None,
    extra_loss_fns: Optional[Dict[str, Callable[[StepData],
                                                jnp.ndarray]]] = None,
    action_shapes: Dict[str, Dict[str, Any]] = None,
    agent_name: str = None,
):
  """Computes PPO loss."""
  policy_params, value_params = models['policy'], models['value']
  extra_params = models.get('extra', {})
  policy_logits = policy_apply(policy_params, data.obs[:-1])
  baseline = value_apply(value_params, data.obs)
  baseline = jnp.squeeze(baseline, axis=-1)

  # Use last baseline value (from the value function) to bootstrap.
  bootstrap_value = baseline[-1]
  baseline = baseline[:-1]

  # At this point, we have unroll length + 1 steps. The last step is only used
  # as bootstrap value, so it's removed.

  # already removed at data generation time
  # actions = actions[:-1]
  # logits = logits[:-1]

  agent_index = list(action_shapes.keys()).index(agent_name)
  rewards = data.rewards[1:, ..., agent_index] * reward_scaling
  truncation = data.truncation[1:]
  termination = data.dones[1:] * (1 - truncation)
  actions = data.actions[agent_index]
  logits = data.logits[agent_index]

  target_action_log_probs = parametric_action_distribution.log_prob(
      policy_logits, actions)
  behaviour_action_log_probs = parametric_action_distribution.log_prob(
      logits, actions)

  vs, advantages = ppo_losses.compute_gae(
      truncation=truncation,
      termination=termination,
      rewards=rewards,
      values=baseline,
      bootstrap_value=bootstrap_value,
      lambda_=lambda_,
      discount=discounting)
  rho_s = jnp.exp(target_action_log_probs - behaviour_action_log_probs)

  surrogate_loss1 = rho_s * advantages
  surrogate_loss2 = jnp.clip(rho_s, 1 - ppo_epsilon,
                             1 + ppo_epsilon) * advantages

  policy_loss = -jnp.mean(jnp.minimum(surrogate_loss1, surrogate_loss2))

  # Value function loss
  v_error = vs - baseline
  value_loss = jnp.mean(v_error * v_error) * 0.5 * 0.5

  # Entropy reward
  entropy = jnp.mean(parametric_action_distribution.entropy(policy_logits, rng))
  entropy_loss = entropy_cost * -entropy

  total_loss = policy_loss + value_loss + entropy_loss

  # Additional losses
  extra_losses = {}
  if extra_loss_fns:
    for key, loss_fn in extra_loss_fns.items():
      loss, rng = loss_fn(data=data, udata=udata, rng=rng, params=extra_params)
      if extra_loss_update_ratios and key in extra_loss_update_ratios:
        # enable loss gradient p*100 percent of the time
        rng, key_update = jax.random.split(rng)
        p = extra_loss_update_ratios[key]
        b = jax.random.bernoulli(key_update, p=jnp.array(p))
        loss = jnp.where(b, loss, jax.lax.stop_gradient(loss))
      total_loss += loss
      extra_losses[key] = loss

  return total_loss, dict(
      extra_losses, **{
          'total_loss': total_loss,
          'policy_loss': policy_loss,
          'value_loss': value_loss,
          'entropy_loss': entropy_loss,
      })


def train(environment_fn: Callable[..., envs.Env],
          num_timesteps,
          episode_length: int,
          action_repeat: int = 1,
          num_envs: int = 1,
          max_devices_per_host: Optional[int] = None,
          num_eval_envs: int = 128,
          learning_rate=1e-4,
          entropy_cost=1e-4,
          discounting=0.9,
          seed=0,
          unroll_length=10,
          batch_size=32,
          num_minibatches=16,
          num_update_epochs=2,
          log_frequency=10,
          normalize_observations=False,
          reward_scaling=1.,
          progress_fn: Optional[Callable[[int, Dict[str, Any]], None]] = None,
          parametric_action_distribution_fn: Optional[Callable[[
              int,
          ], distribution.ParametricDistribution]] = distribution
          .NormalTanhDistribution,
          make_models_fn: Optional[Callable[
              [int, int],
              Tuple[networks.FeedForwardNetwork]]] = networks.make_models,
          policy_params: Optional[Dict[str, jnp.ndarray]] = None,
          value_params: Optional[Dict[str, jnp.ndarray]] = None,
          extra_params: Optional[Dict[str, Dict[str, jnp.ndarray]]] = None,
          extra_step_kwargs: bool = True,
          extra_loss_update_ratios: Optional[Dict[str, float]] = None,
          extra_loss_fns: Optional[Dict[str, Callable[[StepData],
                                                      jnp.ndarray]]] = None):
  """PPO training."""
  assert batch_size * num_minibatches % num_envs == 0
  xt = time.time()

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d', jax.device_count(), process_count,
      process_id, local_device_count, local_devices_to_use)

  # TODO: check key randomness
  key = jax.random.PRNGKey(seed)
  key, key_models, key_env, key_eval = jax.random.split(key, 4)
  # Make sure every process gets a different random key, otherwise they will be
  # doing identical work.
  key_env = jax.random.split(key_env, process_count)[process_id]
  key = jax.random.split(key, process_count)[process_id]
  # key_models should be the same, so that models are initialized the same way
  # for different processes

  key_envs = jax.random.split(key_env, local_devices_to_use)
  core_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_envs // local_devices_to_use // process_count,
      episode_length=episode_length)

  component_env = core_env.unwrapped
  action_shapes = component_env.group_action_shapes
  action_size = component_env.action_size

  core_eval_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_eval_envs,
      episode_length=episode_length)
  eval_first_state, eval_step_fn = env.wrap(
      core_eval_env, key_eval, extra_step_kwargs=extra_step_kwargs)

  tmp_env_states = []
  for key in key_envs:
    first_state, step_fn = env.wrap(
        core_env, key, extra_step_kwargs=extra_step_kwargs)
    tmp_env_states.append(first_state)
  first_state = jax.tree_util.tree_map(lambda *args: jnp.stack(args),
                                  *tmp_env_states)

  normalizer_params, obs_normalizer_update_fn, obs_normalizer_apply_fn = (
      normalization.create_observation_normalizer(
          core_env.observation_size,
          normalize_observations,
          num_leading_batch_dims=2,
          pmap_to_devices=local_devices_to_use))

  agents = odict()
  policy_params = policy_params or [None] * len(action_shapes)
  value_params = value_params or [None] * len(action_shapes)
  for i, (k, action_shape) in enumerate(action_shapes.items()):
    parametric_action_distribution = parametric_action_distribution_fn(
        event_size=action_shape['size'])

    policy_model, value_model = make_models_fn(
        parametric_action_distribution.param_size, core_env.observation_size)
    key_policy, key_value, key_models = jax.random.split(key_models, 3)

    optimizer = optax.adam(learning_rate=learning_rate)
    init_params = {
        'policy': policy_params[i] or policy_model.init(key_policy),
        'value': value_params[i] or value_model.init(key_value),
        'extra': extra_params
    }
    optimizer_state = optimizer.init(init_params)
    optimizer_state, init_params = pmap.bcast_local_devices(
        (optimizer_state, init_params), local_devices_to_use)

    loss_fn = functools.partial(
        compute_ppo_loss,
        parametric_action_distribution=parametric_action_distribution,
        policy_apply=policy_model.apply,
        value_apply=value_model.apply,
        entropy_cost=entropy_cost,
        discounting=discounting,
        reward_scaling=reward_scaling,
        extra_loss_update_ratios=extra_loss_update_ratios,
        extra_loss_fns=extra_loss_fns,
        action_shapes=action_shapes,
        agent_name=k)

    grad_loss = jax.grad(loss_fn, has_aux=True)
    agents[k] = Agent(parametric_action_distribution, policy_model,
                      optimizer_state, init_params, grad_loss)

  key_debug = jax.random.PRNGKey(seed + 666)

  def do_one_step_eval(carry, unused_target_t):
    state, policy_params, normalizer_params, extra_params, key = carry
    key, key_sample = jax.random.split(key)
    obs = obs_normalizer_apply_fn(normalizer_params, state.core.obs)
    actions = odict()
    for i, (k, agent) in enumerate(agents.items()):
      logits = agent.policy_model.apply(policy_params[i], obs)
      actions[k] = agent.parametric_action_distribution.sample(
          logits, key_sample)
    actions_arr = jnp.zeros(obs.shape[:-1] + (action_size,))
    actions = data_utils.fill_array(actions, actions_arr, action_shapes)
    nstate = eval_step_fn(state, actions, normalizer_params, extra_params)
    return (nstate, policy_params, normalizer_params, extra_params, key), ()

  @jax.jit
  def run_eval(state, key, policy_params, normalizer_params, extra_params):
    policy_params = jax.tree_util.tree_map(lambda x: x[0], policy_params)
    normalizer_params = jax.tree_util.tree_map(lambda x: x[0], normalizer_params)
    extra_params = jax.tree_util.tree_map(lambda x: x[0], extra_params)
    (state, _, _, _, key), _ = jax.lax.scan(
        do_one_step_eval,
        (state, policy_params, normalizer_params, extra_params, key), (),
        length=episode_length // action_repeat)
    return state, key

  def do_one_step(carry, unused_target_t):
    state, normalizer_params, policy_params, extra_params, key = carry
    key, key_sample = jax.random.split(key)
    normalized_obs = obs_normalizer_apply_fn(normalizer_params, state.core.obs)
    logits, actions, postprocessed_actions = [], [], odict()
    for i, (k, agent) in enumerate(agents.items()):
      logits += [agent.policy_model.apply(policy_params[i], normalized_obs)]
      actions += [
          agent.parametric_action_distribution.sample_no_postprocessing(
              logits[-1], key_sample)
      ]
      postprocessed_actions[
          k] = agent.parametric_action_distribution.postprocess(actions[-1])
    postprocessed_actions_arr = jnp.zeros(normalized_obs.shape[:-1] +
                                          (action_size,))
    postprocessed_actions = data_utils.fill_array(postprocessed_actions,
                                                  postprocessed_actions_arr,
                                                  action_shapes)
    nstate = step_fn(state, postprocessed_actions, normalizer_params,
                     extra_params)
    return (nstate, normalizer_params, policy_params, extra_params,
            key), StepData(
                obs=state.core.obs,
                rewards=state.core.reward,
                dones=state.core.done,
                truncation=state.core.info['truncation'],
                actions=actions,
                logits=logits)

  def generate_unroll(carry, unused_target_t):
    state, normalizer_params, policy_params, extra_params, key = carry
    (state, _, _, _, key), data = jax.lax.scan(
        do_one_step,
        (state, normalizer_params, policy_params, extra_params, key), (),
        length=unroll_length)
    data = data.replace(
        obs=jnp.concatenate([data.obs,
                             jnp.expand_dims(state.core.obs, axis=0)]),
        rewards=jnp.concatenate(
            [data.rewards,
             jnp.expand_dims(state.core.reward, axis=0)]),
        dones=jnp.concatenate(
            [data.dones, jnp.expand_dims(state.core.done, axis=0)]),
        truncation=jnp.concatenate([
            data.truncation,
            jnp.expand_dims(state.core.info['truncation'], axis=0)
        ]))
    return (state, normalizer_params, policy_params, extra_params, key), data

  def update_model(carry, data_tuple):
    optimizer_state, params, key = carry
    data, udata = data_tuple
    key, key_loss = jax.random.split(key)
    metrics = []
    for i, agent in enumerate(agents.values()):
      loss_grad, agent_metrics = agent.grad_loss(params[i], data, udata,
                                                 key_loss)
      metrics.append(agent_metrics)
      loss_grad = jax.lax.pmean(loss_grad, axis_name='i')
      params_update, optimizer_state[i] = optimizer.update(
          loss_grad, optimizer_state[i])
      params[i] = optax.apply_updates(params[i], params_update)

    return (optimizer_state, params, key), metrics

  def minimize_epoch(carry, unused_t):
    optimizer_state, params, data, udata, key = carry
    key, key_perm, key_grad = jax.random.split(key, 3)
    permutation = jax.random.permutation(key_perm, data.obs.shape[1])

    def convert_data(data, permutation):
      data = jnp.take(data, permutation, axis=1, mode='clip')
      data = jnp.reshape(data, [data.shape[0], num_minibatches, -1] +
                         list(data.shape[2:]))
      data = jnp.swapaxes(data, 0, 1)
      return data

    ndata = jax.tree_util.tree_map(lambda x: convert_data(x, permutation), data)
    u_ndata = jax.tree_util.tree_map(lambda x: convert_data(x, permutation), udata)
    (optimizer_state, params, _), metrics = jax.lax.scan(
        update_model, (optimizer_state, params, key_grad), (ndata, u_ndata),
        length=num_minibatches)
    return (optimizer_state, params, data, udata, key), metrics

  def get_params(state, key, value=None):
    if value is not None:
      return [params.get(key, value) for params in state.params]
    return [params[key] for params in state.params]

  def run_epoch(carry, unused_t):
    training_state, state = carry
    key_minimize, key_generate_unroll, new_key = jax.random.split(
        training_state.key, 3)
    (state, _, _, _, _), data = jax.lax.scan(
        generate_unroll,
        (state, training_state.normalizer_params,
         get_params(training_state, 'policy'),
         get_params(training_state, 'extra', {}), key_generate_unroll), (),
        length=batch_size * num_minibatches // num_envs)
    # make unroll first
    data = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 0, 1), data)
    data = jax.tree_util.tree_map(
        lambda x: jnp.reshape(x, [x.shape[0], -1] + list(x.shape[3:])), data)

    # Update normalization params and normalize observations.
    normalizer_params = obs_normalizer_update_fn(
        training_state.normalizer_params, data.obs[:-1])
    udata = data
    data = data.replace(
        obs=obs_normalizer_apply_fn(normalizer_params, data.obs))

    (optimizer_state, params, _, _, _), metrics = jax.lax.scan(
        minimize_epoch, (training_state.optimizer_state, training_state.params,
                         data, udata, key_minimize), (),
        length=num_update_epochs)

    new_training_state = TrainingState(
        optimizer_state=optimizer_state,
        params=params,
        normalizer_params=normalizer_params,
        key=new_key)
    return (new_training_state, state), metrics

  num_epochs = num_timesteps // (
      batch_size * unroll_length * num_minibatches * action_repeat)

  def _minimize_loop(training_state, state):
    (training_state, state), losses = jax.lax.scan(
        run_epoch, (training_state, state), (),
        length=num_epochs // log_frequency)
    losses = jax.tree_util.tree_map(jnp.mean, losses)
    return (training_state, state), losses

  minimize_loop = jax.pmap(_minimize_loop, axis_name='i')

  inference = make_inference_fn(
      core_env.observation_size, action_shapes, normalize_observations,
      parametric_action_distribution_fn, make_models_fn)

  training_state = TrainingState(
      optimizer_state=[agent.optimizer_state for agent in agents.values()],
      params=[agent.init_params for agent in agents.values()],
      key=jnp.stack(jax.random.split(key, local_devices_to_use)),
      normalizer_params=normalizer_params)
  training_walltime = 0
  eval_walltime = 0
  sps = 0
  eval_sps = 0
  losses = []
  state = first_state
  metrics = {}

  for it in range(log_frequency + 1):
    logging.info('starting iteration %s %s', it, time.time() - xt)
    t = time.time()

    if process_id == 0:
      eval_state, key_debug = (
          run_eval(eval_first_state, key_debug,
                   get_params(training_state, 'policy'),
                   training_state.normalizer_params,
                   get_params(training_state, 'extra', {})))
      eval_state.total_episodes.block_until_ready()
      eval_walltime += time.time() - t
      eval_sps = (
          episode_length * eval_first_state.core.reward.shape[0] /
          (time.time() - t))

      metrics = dict(
          **dict({
              f'eval/episode_{name}': value / eval_state.total_episodes
              for name, value in eval_state.total_metrics.items()
          }),
          **dict({
              f'{index}/losses/{k}': jnp.mean(v)
              for index, loss in enumerate(losses) for k, v in loss.items()
          }),
          **dict({
              'eval/total_episodes': eval_state.total_episodes,
              'speed/sps': sps,
              'speed/eval_sps': eval_sps,
              'speed/training_walltime': training_walltime,
              'speed/eval_walltime': eval_walltime,
              'speed/timestamp': training_walltime
          }))

      logging.info(metrics)
      if progress_fn:
        params = dict(
            normalizer=jax.tree_util.tree_map(lambda x: x[0],
                                    training_state.normalizer_params),
            policy=jax.tree_util.tree_map(lambda x: x[0],
                                get_params(training_state, 'policy')),
            extra=jax.tree_util.tree_map(lambda x: x[0],
                               get_params(training_state, 'extra')))
        progress_fn(
            int(training_state.normalizer_params[0][0]) * action_repeat,
            metrics, params)

    if it == log_frequency:
      break

    t = time.time()
    previous_step = training_state.normalizer_params[0][0]
    # optimization
    (training_state, state), losses = minimize_loop(training_state, state)
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), losses)
    sps = ((training_state.normalizer_params[0][0] - previous_step) /
           (time.time() - t)) * action_repeat
    training_walltime += time.time() - t

  # To undo the pmap.
  normalizer_params = jax.tree_util.tree_map(lambda x: x[0],
                                   training_state.normalizer_params)
  policy_params = jax.tree_util.tree_map(lambda x: x[0],
                               get_params(training_state, 'policy'))
  extra_params = jax.tree_util.tree_map(lambda x: x[0],
                              get_params(training_state, 'extra'))

  logging.info('total steps: %s', normalizer_params[0] * action_repeat)

  params = dict(
      normalizer=normalizer_params, policy=policy_params, extra=extra_params)

  if process_count > 1:
    # Make sure all processes stay up until the end of main.
    x = jnp.ones([jax.local_device_count()])
    x = jax.device_get(jax.pmap(lambda x: jax.lax.psum(x, 'i'), 'i')(x))
    assert x[0] == jax.device_count()

  return (inference, params, metrics)


def make_inference_fn(
    observation_size: int,
    action_shapes: Dict[str, Any],
    normalize_observations: bool = False,
    parametric_action_distribution_fn: Optional[Callable[[
        int,
    ], distribution.ParametricDistribution]] = distribution
    .NormalTanhDistribution,
    make_models_fn: Optional[Callable[
        [int, int], Tuple[networks.FeedForwardNetwork]]] = networks.make_models):
  """Creates params and inference function for the multi-agent PPO agent."""
  action_size = sum([s['size'] for s in action_shapes.values()])
  _, obs_normalizer_apply_fn = normalization.make_data_and_apply_fn(
      observation_size, normalize_observations=normalize_observations)
  agents = odict()
  for k, action_shape in action_shapes.items():
    parametric_action_distribution = parametric_action_distribution_fn(
        event_size=action_shape['size'])
    policy_model, _ = make_models_fn(parametric_action_distribution.param_size,
                                     observation_size)
    agents[k] = (parametric_action_distribution, policy_model)

  def inference_fn(params, obs, key):
    normalizer_params, policy_params = params['normalizer'], params['policy']
    obs = obs_normalizer_apply_fn(normalizer_params, obs)
    actions = odict()
    for i, (k, (parametric_action_distribution,
                policy_model)) in enumerate(agents.items()):
      actions[k] = parametric_action_distribution.sample(
          policy_model.apply(policy_params[i], obs), key)
    actions_arr = jnp.zeros(obs.shape[:-1] + (action_size,))
    actions = data_utils.fill_array(actions, actions_arr, action_shapes)
    return actions

  return inference_fn



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Spring perf tests."""

from absl.testing import absltest
from brax import test_utils
from brax.spring import pipeline
import jax
from jax import numpy as jp


class PerfTest(absltest.TestCase):

  def test_pipeline_ant(self):
    sys = test_utils.load_fixture('ant.xml')

    def init_fn(rng):
      rng1, rng2 = jax.random.split(rng, 2)
      q = jax.random.uniform(rng1, (sys.q_size(),), minval=-0.1, maxval=0.1)
      qd = 0.1 * jax.random.normal(rng2, (sys.qd_size(),))
      return pipeline.init(sys, q, qd)

    def step_fn(state):
      return pipeline.step(sys, state, jp.zeros(sys.act_size()))

    test_utils.benchmark('spring pipeline ant', init_fn, step_fn)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Brax training types."""

from typing import Any, Mapping, NamedTuple, Tuple, TypeVar

from brax.training.acme.types import NestedArray
import jax.numpy as jnp

# Protocol was introduced into typing in Python >=3.8
# via https://www.python.org/dev/peps/pep-0544/
# Before that, its status was DRAFT and available via typing_extensions
try:
  from typing import Protocol  # pylint:disable=g-import-not-at-top
except ImportError:
  from typing_extensions import Protocol  # pylint:disable=g-import-not-at-top

Params = Any
PRNGKey = jnp.ndarray
Metrics = Mapping[str, jnp.ndarray]
Observation = jnp.ndarray
Action = jnp.ndarray
Extra = Mapping[str, Any]
PolicyParams = Any
PreprocessorParams = Any
PolicyParams = Tuple[PreprocessorParams, Params]
NetworkType = TypeVar('NetworkType')


class Transition(NamedTuple):
  """Container for a transition."""
  observation: NestedArray
  action: NestedArray
  reward: NestedArray
  discount: NestedArray
  next_observation: NestedArray
  extras: NestedArray = ()  # pytype: disable=annotation-type-mismatch  # jax-ndarray


class Policy(Protocol):

  def __call__(
      self,
      observation: Observation,
      key: PRNGKey,
  ) -> Tuple[Action, Extra]:
    pass


class PreprocessObservationFn(Protocol):

  def __call__(
      self,
      observation: Observation,
      preprocessor_params: PreprocessorParams,
  ) -> jnp.ndarray:
    pass


def identity_observation_preprocessor(observation: Observation,
                                      preprocessor_params: PreprocessorParams):
  del preprocessor_params
  return observation


class NetworkFactory(Protocol[NetworkType]):

  def __call__(
      self,
      observation_size: int,
      action_size: int,
      preprocess_observations_fn:
      PreprocessObservationFn = identity_observation_preprocessor
  ) -> NetworkType:
    pass



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Physics pipeline for generalized coordinates engine."""

from typing import Optional
from brax import actuator
from brax import contact
from brax import kinematics
from brax.base import System
from brax.generalized import constraint
from brax.generalized import dynamics
from brax.generalized import integrator
from brax.generalized import mass
from brax.generalized.base import State
from brax.io import mjcf
import jax


def init(
    sys: System,
    q: jax.Array,
    qd: jax.Array,
    unused_act: Optional[jax.Array] = None,
    debug: bool = False,
) -> State:
  """Initializes physics state.

  Args:
    sys: a brax system
    q: (q_size,) joint angle vector
    qd: (qd_size,) joint velocity vector
    debug: if True, adds contact to the state for debugging

  Returns:
    state: initial physics state
  """
  if sys.mj_model is not None:
    mjcf.validate_model(sys.mj_model)
  x, xd = kinematics.forward(sys, q, qd)
  state = State.init(q, qd, x, xd)  # pytype: disable=wrong-arg-types  # jax-ndarray
  state = dynamics.transform_com(sys, state)
  state = mass.matrix_inv(sys, state, 0)
  state = constraint.jacobian(sys, state)
  if debug:
    state = state.replace(contact=contact.get(sys, state.x))

  return state


def step(
    sys: System, state: State, act: jax.Array, debug: bool = False
) -> State:
  """Performs a physics step.

  Args:
    sys: a brax system
    state: physics state prior to step
    act: (act_size,) actuator input vector
    debug: if True, adds contact to the state for debugging

  Returns:
    state: physics state after step
  """
  # calculate acceleration terms
  tau = actuator.to_tau(sys, act, state.q, state.qd)
  state = state.replace(qf_smooth=dynamics.forward(sys, state, tau))
  state = state.replace(qf_constraint=constraint.force(sys, state))

  # update position/velocity level terms
  state = integrator.integrate(sys, state)
  x, xd = kinematics.forward(sys, state.q, state.qd)
  state = state.replace(x=x, xd=xd)
  state = dynamics.transform_com(sys, state)
  state = mass.matrix_inv(sys, state, sys.matrix_inv_iterations)
  state = constraint.jacobian(sys, state)

  if debug:
    state = state.replace(contact=contact.get(sys, state.x))

  return state



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Brax training acting functions."""

import time
from typing import Callable, Sequence, Tuple, Union

from brax import envs
from brax.training.types import Metrics
from brax.training.types import Policy
from brax.training.types import PolicyParams
from brax.training.types import PRNGKey
from brax.training.types import Transition
from brax.v1 import envs as envs_v1
import jax
import numpy as np

State = Union[envs.State, envs_v1.State]
Env = Union[envs.Env, envs_v1.Env, envs_v1.Wrapper]


def actor_step(
    env: Env,
    env_state: State,
    policy: Policy,
    key: PRNGKey,
    extra_fields: Sequence[str] = ()
) -> Tuple[State, Transition]:
  """Collect data."""
  actions, policy_extras = policy(env_state.obs, key)
  nstate = env.step(env_state, actions)
  state_extras = {x: nstate.info[x] for x in extra_fields}
  return nstate, Transition(  # pytype: disable=wrong-arg-types  # jax-ndarray
      observation=env_state.obs,
      action=actions,
      reward=nstate.reward,
      discount=1 - nstate.done,
      next_observation=nstate.obs,
      extras={
          'policy_extras': policy_extras,
          'state_extras': state_extras
      })


def generate_unroll(
    env: Env,
    env_state: State,
    policy: Policy,
    key: PRNGKey,
    unroll_length: int,
    extra_fields: Sequence[str] = ()
) -> Tuple[State, Transition]:
  """Collect trajectories of given unroll_length."""

  @jax.jit
  def f(carry, unused_t):
    state, current_key = carry
    current_key, next_key = jax.random.split(current_key)
    nstate, transition = actor_step(
        env, state, policy, current_key, extra_fields=extra_fields)
    return (nstate, next_key), transition

  (final_state, _), data = jax.lax.scan(
      f, (env_state, key), (), length=unroll_length)
  return final_state, data


# TODO: Consider moving this to its own file.
class Evaluator:
  """Class to run evaluations."""

  def __init__(self, eval_env: envs.Env,
               eval_policy_fn: Callable[[PolicyParams],
                                        Policy], num_eval_envs: int,
               episode_length: int, action_repeat: int, key: PRNGKey):
    """Init.

    Args:
      eval_env: Batched environment to run evals on.
      eval_policy_fn: Function returning the policy from the policy parameters.
      num_eval_envs: Each env will run 1 episode in parallel for each eval.
      episode_length: Maximum length of an episode.
      action_repeat: Number of physics steps per env step.
      key: RNG key.
    """
    self._key = key
    self._eval_walltime = 0.

    eval_env = envs.training.EvalWrapper(eval_env)

    def generate_eval_unroll(policy_params: PolicyParams,
                             key: PRNGKey) -> State:
      reset_keys = jax.random.split(key, num_eval_envs)
      eval_first_state = eval_env.reset(reset_keys)
      return generate_unroll(
          eval_env,
          eval_first_state,
          eval_policy_fn(policy_params),
          key,
          unroll_length=episode_length // action_repeat)[0]

    self._generate_eval_unroll = jax.jit(generate_eval_unroll)
    self._steps_per_unroll = episode_length * num_eval_envs

  def run_evaluation(self,
                     policy_params: PolicyParams,
                     training_metrics: Metrics,
                     aggregate_episodes: bool = True) -> Metrics:
    """Run one epoch of evaluation."""
    self._key, unroll_key = jax.random.split(self._key)

    t = time.time()
    eval_state = self._generate_eval_unroll(policy_params, unroll_key)
    eval_metrics = eval_state.info['eval_metrics']
    eval_metrics.active_episodes.block_until_ready()
    epoch_eval_time = time.time() - t
    metrics = {}
    for fn in [np.mean, np.std]:
      suffix = '_std' if fn == np.std else ''
      metrics.update(
          {
              f'eval/episode_{name}{suffix}': (
                  fn(value) if aggregate_episodes else value
              )
              for name, value in eval_metrics.episode_metrics.items()
          }
      )
    metrics['eval/avg_episode_length'] = np.mean(eval_metrics.episode_steps)
    metrics['eval/epoch_eval_time'] = epoch_eval_time
    metrics['eval/sps'] = self._steps_per_unroll / epoch_eval_time
    self._eval_walltime = self._eval_walltime + epoch_eval_time
    metrics = {
        'eval/walltime': self._eval_walltime,
        **training_metrics,
        **metrics
    }

    return metrics  # pytype: disable=bad-return-type  # jax-ndarray



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Joint definition and apply functions."""
# pylint:disable=g-multiple-import
from typing import Tuple

from brax import com
from brax import kinematics
from brax import math
from brax import scan
from brax.base import DoF, Force, Link, Motion, System, Transform
from brax.positional.base import State
import jax
from jax import numpy as jp
from jax.ops import segment_sum


def acceleration_update(sys: System, state: State, tau: jax.Array) -> Force:
  """Calculates forces to apply to links resulting from joint constraints.

  Args:
    sys: System defining kinematic tree of joints
    state: positional pipeline state
    tau: joint force vector

  Returns:
    xf_i: force to apply to link center of mass in world frame
  """

  def _free_joint(*_) -> Force:
    return Force(vel=jp.zeros(3), ang=jp.zeros(3))

  def _damp(link: Link, jd: Motion, dof: DoF, tau: jax.Array):
    vel = jp.sum(jax.vmap(jp.multiply)(tau, dof.motion.vel), axis=0)
    ang = jp.sum(jax.vmap(jp.multiply)(tau, dof.motion.ang), axis=0)

    # damp the angular and linear motion
    ang -= link.constraint_ang_damping * jd.ang
    vel -= link.constraint_vel_damping * jd.vel

    return Force(ang=ang, vel=vel)

  def j_fn(typ, link, jd, dof, tau):
    # change dof-shape variables into link-shape
    reshape_fn = lambda x: x.reshape((jd.ang.shape[0], -1) + x.shape[1:])
    tau, dof = jax.tree.map(reshape_fn, (tau, dof))
    j_fn_map = {
        'f': _free_joint,
        '1': _damp,
        '2': _damp,
        '3': _damp,
    }

    return jax.vmap(j_fn_map[typ])(link, jd, dof, tau)

  # calculate forces in joint frame, then convert to world frame
  link, jd, dof = sys.link, state.jd, sys.dof
  jf = scan.link_types(sys, j_fn, 'lldd', 'l', link, jd, dof, tau)
  xf = Transform.create(rot=state.a_p.rot).vmap().do(jf)
  # move force to center of mass offset
  fc = Transform.create(pos=state.a_c.pos - state.x_i.pos).vmap().do(xf)
  # also add opposite force to parent link at center of mass
  parent_idx = jp.array(sys.link_parents)
  x_i_parent = state.x_i.take(parent_idx)
  fp = Transform.create(pos=state.a_p.pos - x_i_parent.pos).vmap().do(xf)
  fp = jax.tree.map(lambda x: segment_sum(x, parent_idx, sys.num_links()), fp)
  xf_i = fc - fp
  return xf_i


def position_update(sys: System, state: State) -> Transform:
  """Calculates position-level joint updates in CoM coordinates for joints.

  Args:
    sys: System defining kinematic tree of joints
    state: positional pipeline state

  Returns:
    x_i: new position after update
  """

  p_idx = jp.array(sys.link_parents)
  xi_p = state.x_i.concatenate(Transform.zero((1,))).take(p_idx)
  j, _, a_p, a_c = kinematics.world_to_joint(sys, state.x, state.xd)

  # pad sys and dof data withs 0s along inactive axes
  d_j = jax.vmap(_three_dof_joint_update)(j, *_sphericalize(sys, j))
  free_mask = jp.array([l != 'f' for l in sys.link_types])
  d_j = jax.tree.map(lambda x: jax.vmap(jp.multiply)(x, free_mask), d_j)
  d_w = jax.tree.map(lambda x: jax.vmap(math.rotate)(x, a_p.rot), d_j)

  i_inv = com.inv_inertia(sys, state.x)
  i_inv_p = jax.vmap(jp.multiply)(i_inv[p_idx], p_idx > -1)
  mass_inv = 1 / (sys.link.inertia.mass ** (1 - sys.spring_mass_scale))
  mass_inv_p = mass_inv[p_idx] * (p_idx > -1)
  dp_p_pos, dp_c_pos = jax.vmap(_translation_update)(
      a_p, xi_p, i_inv_p, mass_inv_p, a_c, state.x_i, i_inv, mass_inv, -d_w.pos)
  dp_p_ang, dp_c_ang = jax.vmap(_rotation_update)(
      xi_p, i_inv_p, state.x_i, i_inv, d_w.rot
  )

  dp_c = dp_c_pos * sys.joint_scale_pos + dp_c_ang * sys.joint_scale_ang
  dp_p = dp_p_pos * sys.joint_scale_pos + dp_p_ang * sys.joint_scale_ang
  dp_p = jax.tree.map(lambda f: segment_sum(f, p_idx, sys.num_links()), dp_p)

  return state.x_i + dp_c + dp_p


def _translation_update(
    pos_p: Transform,
    xi_p: Transform,
    i_inv_p: jax.Array,
    mass_inv_p: jax.Array,
    pos_c: Transform,
    xi_c: Transform,
    i_inv_c: jax.Array,
    mass_inv_c: jax.Array,
    dx: jax.Array,
) -> Tuple[Transform, Transform]:
  """Calculates a position based translational update."""

  pos_p, pos_c = pos_p.pos - xi_p.pos, pos_c.pos - xi_c.pos
  n, c = math.normalize(dx)

  cr1, cr2 = jp.cross(pos_p, n), jp.cross(pos_c, n)
  w1 = mass_inv_p + jp.dot(cr1, i_inv_p @ cr1)
  w2 = mass_inv_c + jp.dot(cr2, i_inv_c @ cr2)
  dlambda = -c / (w1 + w2 + 1e-6)
  p = dlambda * n

  rot_p = -0.5 * math.vec_quat_mul(i_inv_p @ jp.cross(pos_p, p), xi_p.rot)
  rot_c = 0.5 * math.vec_quat_mul(i_inv_c @ jp.cross(pos_c, p), xi_c.rot)
  pos_p, pos_c = -p * mass_inv_p, p * mass_inv_c

  return Transform(pos=pos_p, rot=rot_p), Transform(pos=pos_c, rot=rot_c)


def _rotation_update(
    xi_p: Transform,
    i_inv_p: jax.Array,
    xi_c: Transform,
    i_inv_c: jax.Array,
    dq: jax.Array,
) -> Tuple[Transform, Transform]:
  """Calculates a position based rotational update."""

  n, th = math.normalize(dq)

  # ignoring inertial effects for now
  w1, w2 = jp.dot(n, i_inv_p @ n), jp.dot(n, i_inv_c @ n)
  dlambda = -th / (w1 + w2 + 1e-6)
  p = -dlambda * n

  rot_p = -0.5 * math.vec_quat_mul(i_inv_p @ p, xi_p.rot)
  rot_c = 0.5 * math.vec_quat_mul(i_inv_c @ p, xi_c.rot)

  return Transform.create(rot=rot_p), Transform.create(rot=rot_c)


def _sphericalize(sys, j):
  """Transforms system state into an all-3-dof version of the system."""

  def pad_free(_):
    # create dummy data for free links
    inf = jp.array([jp.inf, jp.inf, jp.inf])
    return (-inf, inf), Motion(ang=jp.eye(3), vel=jp.eye(3)),

  def pad_x_dof(dof, x):
    if dof.limit:
      stack_fn = lambda a: jp.concatenate((a, jp.zeros(3 - x)))
      limit = jax.tree.map(stack_fn, dof.limit)
    else:
      inf = jp.array([jp.inf, jp.inf, jp.inf])
      limit = (-inf, inf)
    padded_motion = dof.motion.concatenate(Motion.zero((3 - x,)))
    return limit, padded_motion

  def j_fn(typ, j, dof):
    # change dof-shape variables into link-shape
    reshape_fn = lambda x: x.reshape((j.pos.shape[0], -1) + x.shape[1:])
    dof = jax.tree.map(reshape_fn, dof)
    j_fn_map = {
        'f': pad_free,
        '1': lambda x: pad_x_dof(x, 1),
        '2': lambda x: pad_x_dof(x, 2),
        '3': lambda x: pad_x_dof(x, 3),
    }
    limit, padded_motion = jax.vmap(j_fn_map[typ])(dof)

    if typ == 'f':
      joint_frame_fn = lambda x: (Motion(vel=jp.eye(3), ang=jp.eye(3)), 1)
    else:
      joint_frame_fn = kinematics.link_to_joint_frame
    joint_frame, parity = jax.vmap(joint_frame_fn)(dof.motion)

    return limit, padded_motion, joint_frame, parity

  result = scan.link_types(sys, j_fn, 'ld', 'l', j, sys.dof)

  return result


def _three_dof_joint_update(
    x: Transform,
    limit: Tuple[jax.Array, jax.Array],
    motion: Motion,
    joint_frame: Motion,
    parity: float,
) -> Transform:
  """Returns position-level displacements in joint frame for spherical joint."""
  (
      axis_c,
      (_, _, _),
      (line_of_nodes, axis_1_p_in_xz_c),
  ) = kinematics.axis_angle_ang(x, joint_frame, parity)

  axis_p = joint_frame.ang

  axis_1_p_in_xz_c = (
      jp.dot(axis_p[0], axis_c[0]) * axis_c[0]
      + jp.dot(axis_p[0], axis_c[1]) * axis_c[1]
  )

  axis_1_p_in_xz_c, _ = math.normalize(axis_1_p_in_xz_c)
  axis_2_normal, _ = math.normalize(jp.cross(axis_1_p_in_xz_c, axis_p[0]))
  limit_axes = jp.array([
      axis_p[0],
      -axis_2_normal * jp.sign(jp.dot(axis_p[0], axis_c[2])),
      axis_c[2],
  ])

  ref_axis_1 = jp.array([axis_p[1], axis_p[0], line_of_nodes])
  ref_axis_2 = jp.array([line_of_nodes, axis_1_p_in_xz_c, axis_c[1]])

  def limit_angle(n, n_1, n_2, motion, limit, ang_limit):
    ph = math.signed_angle(n, n_1, n_2)
    ph = jp.clip(ph, ang_limit[0], ang_limit[1])
    fixrot = math.quat_rot_axis(n, ph)
    n1 = math.rotate(n_1, fixrot)
    dq = jp.cross(n1, n_2)

    active_axis = motion.vel.any()
    xp = motion.vel @ x.pos
    dx = xp - jp.clip(xp, limit[0], limit[1])
    dx = motion.vel * dx * active_axis

    return dq, dx

  # positional constraints
  dx = -x.pos

  # remove components of update along free prismatic axes
  dx *= 1 - motion.vel.any(axis=0)

  # limit constraints
  if limit:
    # for freezing angular dofs on prismatic axes
    padded_ang_limit = jp.where(
        motion.vel.any(axis=1),
        jp.zeros((2, 3)),
        jp.array(limit),
    ).transpose()

    dq, dx_lim = jax.vmap(limit_angle)(
        limit_axes,
        ref_axis_1,
        ref_axis_2,
        motion,
        limit,
        padded_ang_limit,
    )
    dq = -1.0 * jp.sum(dq, axis=0)
    dx -= jp.sum(dx_lim, axis=0)
  else:
    dq = jp.zeros_like(x.pos)

  return Transform(pos=dx, rot=dq)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for integrating maximal coordinate dynamics."""
# pylint:disable=g-multiple-import
from typing import Tuple

from brax import math
from brax.base import Motion, System, Transform
import jax
from jax import numpy as jp


def integrate(
    sys: System,
    x_i: Transform,
    xd_i: Motion,
    xdv_i: Motion,
) -> Tuple[Transform, Motion]:
  """Updates state with velocity update in the center of mass frame.

  Args:
    sys: System to forward propagate
    x_i: link center of mass transform in world frame
    xd_i: link center of mass motion in world frame
    xdv_i: link center of mass delta-velocity in world frame

  Returns:
    x_i: updated link center of mass transform in world frame
    xd_i: updated link center of mass motion in world frame
  """

  @jax.vmap
  def op(x_i, xd_i, xdv_i):
    # damp velocity and add acceleration
    xd_i = Motion(
        vel=jp.exp(sys.vel_damping * sys.opt.timestep) * xd_i.vel,
        ang=jp.exp(sys.ang_damping * sys.opt.timestep) * xd_i.ang,
    )
    xd_i += xdv_i

    rot_at_ang_quat = math.ang_to_quat(xd_i.ang) * 0.5 * sys.opt.timestep
    rot = x_i.rot + math.quat_mul(rot_at_ang_quat, x_i.rot)
    x_i = Transform(
        pos=x_i.pos + xd_i.vel * sys.opt.timestep, rot=rot / jp.linalg.norm(rot)
    )

    return x_i, xd_i

  return op(x_i, xd_i, xdv_i)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Component loader and editor."""
# pylint:disable=protected-access
# pylint:disable=g-complex-comprehension
# pylint:disable=g-doc-args
# pylint:disable=g-doc-return-or-yield
import copy
import itertools
import json
from typing import Any, Tuple, Dict

import brax.v1 as brax
# pylint:disable=unused-import
from brax.v1.experimental.composer.components import register_component
from brax.v1.experimental.composer.components import register_default_components
# pylint:enable=unused-import
from google.protobuf import text_format
from google.protobuf.json_format import MessageToJson
from google.protobuf.json_format import Parse

DEFAULT_GLOBAL_OPTIONS_STR = """
friction: 1.0
gravity { z: -9.8 }
angular_damping: -0.05
baumgarte_erp: 0.1
dt: 0.05
substeps: 10
dynamics_mode: "legacy_spring"
"""

FIX_XZ_OPTIONS_STR = """
frozen {
  position {
    y: 1.0
  }
  rotation {
    x: 1.0
    z: 1.0
  }
}
"""

NAME_FIELDS = {
    'bodies': ('name',),
    'joints': (
        'name',
        'parent',
        'child',
    ),
    'actuators': (
        'name',
        'joint',
    ),
    'forces': (
        'name',
        'body',
    ),
    'collide_include': (
        'first',
        'second',
    ),
    'qps': ('name',),
    'angles': ('name',),
}


def message_str2message(message_str: str) -> brax.Config:
  return text_format.Parse(message_str, brax.Config())


def message_str2json(message_str: str) -> Dict[str, Any]:
  message = text_format.Parse(message_str, brax.Config())
  json_str = MessageToJson(message)
  return json.loads(json_str)


DEFAULT_GLOBAL_OPTIONS_JSON = message_str2json(DEFAULT_GLOBAL_OPTIONS_STR)


def json2message_str(config_dict: Dict[str, Any]) -> str:
  json_str = json.dumps(config_dict)
  message = Parse(json_str, brax.Config())
  return text_format.MessageToString(message)


def json_global_options(fix_xz=False, **kwargs):
  message_str = DEFAULT_GLOBAL_OPTIONS_STR
  if fix_xz:
    message_str += FIX_XZ_OPTIONS_STR
  options = message_str2json(message_str)
  return dict(copy.deepcopy(options), **kwargs)


def json_collides(first_collides: Tuple[str],
                  second_collides: Tuple[str]) -> Dict[str, Any]:
  collides = []
  for first, second in itertools.product(first_collides, second_collides):
    collides += [dict(first=first, second=second)]
  return dict(collide_include=collides)


SPLITTER = '___'
COMP_SPLITTER = '__'


def concat_comps(*comp_names):
  """Concat comp_name."""
  if not comp_names:
    return None
  return COMP_SPLITTER.join(sorted(comp_names))


def match_name(name: str, *comp_names):
  """Check comp_name matches name, e.g. ('a1__a2___dist','a1','a2')-->True."""
  comp_name = concat_comps(*comp_names)
  assert comp_name
  return name.startswith(f'{comp_name}{SPLITTER}')


def concat_name(name: str, *comp_names):
  """Add comp_name to name, e.g. ('dist', 'a2', 'a1')-->'a1__a2___dist'."""
  comp_name = concat_comps(*comp_names)
  if comp_name:
    return f'{comp_name}{SPLITTER}{name}'
  return name


def split_name(name: str):
  """Split the name, e.g. ('a1__a2___dist')-->('dist', ('a1', 'a2'))."""
  if SPLITTER not in name:
    return name, []
  comp_name, name = name.split(SPLITTER, 1)
  comp_names = comp_name.split(COMP_SPLITTER)
  return name, comp_names


def json_concat_name(
    config_dict: Any,
    comp_name: str = '',
    parents: Tuple[str] = (),
    excludes: Tuple[str] = ('Ground',),
    force_add: bool = False,
) -> Dict[str, Any]:
  """Add comp_name to all name references in config."""
  if isinstance(config_dict, dict):
    return {
        key: json_concat_name(
            value,
            comp_name=comp_name,
            parents=parents + (key,),
            excludes=excludes,
            force_add=force_add) for key, value in config_dict.items()
    }
  elif isinstance(config_dict, (list, tuple)):
    return type(config_dict)([
        json_concat_name(
            value,
            comp_name=comp_name,
            parents=parents,
            excludes=excludes,
            force_add=force_add) for value in config_dict
    ])
  else:
    if force_add or (len(parents) >= 2 and
                     parents[-1] in NAME_FIELDS.get(parents[-2], ())):
      assert isinstance(config_dict, str), config_dict
      if config_dict not in excludes:
        config_dict = concat_name(config_dict, comp_name)
    return config_dict


def filter_json(config_dict: Dict[str, Any], ground_name: str):
  """Filter config dict and remove ground & global configs."""
  # filter global configs
  config_dict = {
      k: v
      for k, v in config_dict.items()
      if k not in DEFAULT_GLOBAL_OPTIONS_JSON
  }
  # filter ground body and contacts
  collide_include = config_dict.get('collideInclude', None)
  assert collide_include, f'missing collideInclude?: {sorted(config_dict)}'
  collide_include = [
      c for c in collide_include
      if not any(v == ground_name for v in c.values())
  ]
  config_dict['collide_include'] = collide_include
  bodies = config_dict.get('bodies', None)
  assert bodies, f'missing bodies?: {sorted(config_dict)}'
  bodies = [c for c in bodies if c['name'] != ground_name]
  config_dict['bodies'] = bodies
  return config_dict


def filter_message_str(config_str: str, *args, **kwargs):
  """Filter config str and remove ground & global configs."""
  config_dict = message_str2json(config_str)
  config_dict = filter_json(config_dict, *args, **kwargs)
  return json2message_str(config_dict)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for spring physics pipeline."""

from absl.testing import absltest
from brax import test_utils
from brax.mjx import pipeline
import jax
from jax import numpy as jp
import mujoco
import numpy as np


class PipelineTest(absltest.TestCase):

  def test_pendulum(self):
    model = test_utils.load_fixture('double_pendulum.xml')

    state = pipeline.init(model, model.init_q, jp.zeros(model.qd_size()))
    step_fn = jax.jit(pipeline.step)
    for _ in range(20):
      state = step_fn(model, state, jp.zeros(model.act_size()))

    # compare against mujoco
    model = test_utils.load_fixture_mujoco('double_pendulum.xml')
    data = mujoco.MjData(model)
    mujoco.mj_step(model, data, 20)

    np.testing.assert_almost_equal(data.qpos, state.q, decimal=4)
    np.testing.assert_almost_equal(data.qvel, state.qd, decimal=3)
    np.testing.assert_almost_equal(data.xpos[1:], state.x.pos, decimal=4)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluators."""

from typing import Dict, Tuple

from brax.v1.experimental.braxlines.common import evaluators
from brax.v1.experimental.braxlines.irl_smm import utils as irl_utils
from brax.v1.io import file
import jax
from jax import numpy as jnp
import matplotlib.pyplot as plt


def visualize_disc(
    params: Dict[str, Dict[str, jnp.ndarray]],
    disc: irl_utils.IRLDiscriminator,
    num_grid=25,
    fig=None,
    axs=None,
    figsize: Tuple[float] = (3.5, 3),
    output_path: str = None,
    output_name: str = 'smm',
):
  """Visualize discriminator."""
  xgrid = jnp.linspace(-disc.obs_scale_2d[0], disc.obs_scale_2d[0], num_grid)
  ygrid = jnp.linspace(-disc.obs_scale_2d[1], disc.obs_scale_2d[1], num_grid)
  xgrid, ygrid = jnp.meshgrid(xgrid, ygrid)
  datagrid = jnp.concatenate(
      [xgrid.reshape(-1, 1), ygrid.reshape(-1, 1)], axis=-1)

  # plot discriminator visualization
  distgrid = disc.dist(
      datagrid[..., :len(disc.obs_indices)], params=params['extra'])
  probsgrid = jax.nn.sigmoid(distgrid.logits)
  colors = jnp.clip(jnp.array([[-2, 0, 2]]) * (probsgrid - 0.5), a_min=0)
  if fig is None or axs is None:
    fig, ax = plt.subplots(ncols=1, figsize=figsize)
    axs = [ax]
  axs[-1].scatter(x=datagrid[:, 0], y=datagrid[:, 1], c=colors)
  axs[-1].set_xlim([-disc.obs_scale_2d[0], disc.obs_scale_2d[0]])
  axs[-1].set_ylim([-disc.obs_scale_2d[1], disc.obs_scale_2d[1]])
  axs[-1].set(xlabel=disc.obs_labels_2d[0], ylabel=disc.obs_labels_2d[1])
  axs[-1].set(title='discriminator output (red=0, black=0.5, blue=1)')
  fig.tight_layout()
  if output_path:
    file.MakeDirs(output_path)
    with file.File(f'{output_path}/{output_name}.png', 'wb') as f:
      plt.savefig(f)


def estimate_energy_distance_metric(
    params: Dict[str, Dict[str, jnp.ndarray]],
    disc: irl_utils.IRLDiscriminator,
    target_data: jnp.ndarray,
    env_fn,
    inference_fn=None,
    num_samples: int = 10,
    time_subsampling: int = 10,
    time_last_n: int = 500,
    seed: int = 0,
    visualize: bool = False,
    figsize: Tuple[float] = (3.5, 3),
    output_path: str = None,
    output_name: str = 'smm_dist',
):
  """Estimate energy distance and (optimally) visualize smm."""
  batch_env, states = evaluators.rollout_env(
      env_fn=env_fn,
      params=params,
      batch_size=num_samples,
      inference_fn=inference_fn,
      step_args=(params['normalizer'], params['extra']),
      seed=seed,
  )
  # M = batch_size
  I = disc.indexed_obs_size

  # Get indices of interest
  obses_full = jnp.stack([state.obs for state in states], axis=0)
  obses = obses_full[-time_last_n:][::time_subsampling]
  env_vars = batch_env.disc.index_obs(obses)  # [T, M, I]
  target_vars = target_data
  env_vars_flat = env_vars.reshape(-1, I)
  target_vars_flat = target_vars.reshape(-1, I)

  # Compute energy distance
  norm = jnp.linalg.norm
  ee = norm(env_vars_flat[None] - env_vars_flat[:, None], axis=-1).mean()
  tt = norm(target_vars_flat[None] - target_vars_flat[:, None], axis=-1).mean()
  et = norm(target_vars_flat[None] - env_vars_flat[:, None], axis=-1).mean()
  energy_dist = 2 * et - ee - tt
  metrics = {'metrics/energy_dist': energy_dist}

  # Plot
  if visualize:
    env_vars = irl_utils.make_2d(env_vars)
    target_vars = irl_utils.make_2d(target_vars)
    fig, axs = plt.subplots(ncols=2, figsize=(figsize[0] * 2, figsize[1]))
    axs[0].set(title='agent policy')
    axs[0].set_xlim([-disc.obs_scale_2d[0], disc.obs_scale_2d[0]])
    axs[0].set_ylim([-disc.obs_scale_2d[1], disc.obs_scale_2d[1]])
    axs[0].set(xlabel=disc.obs_labels_2d[0], ylabel=disc.obs_labels_2d[1])
    axs[0].scatter(
        x=env_vars_flat[:, 0], y=env_vars_flat[:, 1], c=[1, 0, 0], alpha=0.3)
    axs[1].set(title='target')
    axs[1].set_xlim([-disc.obs_scale_2d[0], disc.obs_scale_2d[0]])
    axs[1].set_ylim([-disc.obs_scale_2d[1], disc.obs_scale_2d[1]])
    axs[1].set(xlabel=disc.obs_labels_2d[0], ylabel=disc.obs_labels_2d[1])
    axs[1].scatter(
        x=target_vars_flat[:, 0],
        y=target_vars_flat[:, 1],
        c=[0, 0, 1],
        alpha=0.3)
    fig.tight_layout()

    if output_path:
      file.MakeDirs(output_path)
      with file.File(f'{output_path}/{output_name}.png', 'wb') as f:
        plt.savefig(f)

  return metrics



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Composer tests."""

import collections
import functools
from absl.testing import absltest
from absl.testing import parameterized
from brax.v1.experimental.composer import composer
from brax.v1.experimental.composer import data_utils
import jax
from jax import numpy as jnp


class ComposerTest(parameterized.TestCase):
  """Tests for Composer module."""

  @parameterized.parameters('ant_push', 'ant_chase', 'pro_ant_run')
  def testEnvCreation(self, env_name):
    env = composer.create(env_name=env_name)
    env.reset(rng=jax.random.PRNGKey(0))

  @parameterized.parameters('chase', 'follow')
  def testMultiAgentEnvCreation(self, env_name):
    env = composer.create(env_name=env_name)
    state = env.reset(rng=jax.random.PRNGKey(0))
    assert state.reward.ndim > 0, state.reward

  def testInspect(self):
    env_params, support_kwargs = composer.inspect_env('pro_ant_run')
    assert tuple(sorted(env_params)) == ('num_legs',), env_params
    assert not support_kwargs, support_kwargs

  def testActionConcatSplit(self):
    env = composer.create(env_name='humanoid')
    action_shapes = composer.get_action_shapes(env.sys)
    leading_dims = (5,)
    actions = jnp.zeros(leading_dims + (env.action_size,))
    actions_dict = data_utils.split_array(actions, action_shapes)
    actions2 = data_utils.concat_array(actions_dict, action_shapes)
    assert actions.shape == actions2.shape, f'{actions.shape} != {actions2.shape}'

  def testObservationConcatSplit(self):
    leading_dims = (5, 4)
    obs_shapes = ((3, 4), (12,), (2, 2))
    obs_sizes = [
        functools.reduce(lambda x, y: x * y, shape) for shape in obs_shapes
    ]
    obs_vec_size = sum(obs_sizes)
    obs_dict = collections.OrderedDict([(i, jnp.zeros(leading_dims + shape))
                                        for i, shape in enumerate(obs_shapes)])
    # get observer_shapes
    obs_shapes_from_data = data_utils.get_array_shapes(
        obs_dict, batch_shape=leading_dims)
    for s1, s2 in zip(obs_shapes, obs_shapes_from_data.values()):
      s2 = s2['shape']
      self.assertEqual(s1, s2, f'{s1} != {s2}')

    # concat
    obs = data_utils.concat_array(obs_dict, obs_shapes_from_data)
    self.assertEqual(obs.shape, leading_dims + (obs_vec_size,),
                     f'{obs.shape} != {leading_dims} + ({obs_vec_size},)')

    # split again
    obs_dict_2 = data_utils.split_array(obs, obs_shapes_from_data)
    for s1, s2 in zip(obs_dict_2.values(), obs_shapes_from_data.values()):
      s1 = s1.shape
      s2 = s2['shape']
      self.assertEqual(s1, leading_dims + s2, f'{s1} != {leading_dims} + {s2}')


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for scan functions."""

from absl.testing import absltest
from brax import scan
from brax import test_utils
import numpy as np


class ScanTest(absltest.TestCase):

  def test_tree(self):
    """Test scanning down 3 tree levels of ant."""
    sys = test_utils.load_fixture('ant.xml')

    qs, qds = [], []

    def f(parent, link, q, qd):
      if parent is not None:
        link += parent
      qs.append(q)
      qds.append(qd)
      return link

    x = scan.tree(
        sys,
        f,
        'lqd',
        np.arange(sys.num_links()),
        np.arange(sys.q_size()),
        np.arange(sys.qd_size()),
    )

    # our scan carry is the sum of index + ancestry indices for each link
    np.testing.assert_array_equal(
        x,
        np.array([
            0,  # 0
            1,  # 1 + 0
            3,  # 2 + 1 + 0
            3,  # 3 + 0
            7,  # 4 + 3 + 0
            5,  # 5 + 0
            11,  # 6 + 5 + 0
            7,  # 7 + 0
            15,  # 8 + 7 + 0
        ]),
    )

    self.assertLen(qs, 3)  # three levels
    np.testing.assert_array_equal(qs[0], np.array([0, 1, 2, 3, 4, 5, 6]))
    np.testing.assert_array_equal(qs[1], np.array([7, 9, 11, 13]))
    np.testing.assert_array_equal(qs[2], np.array([8, 10, 12, 14]))

    self.assertLen(qds, 3)  # three levels
    np.testing.assert_array_equal(qds[0], np.array([0, 1, 2, 3, 4, 5]))
    np.testing.assert_array_equal(qds[1], np.array([6, 8, 10, 12]))
    np.testing.assert_array_equal(qds[2], np.array([7, 9, 11, 13]))

  def test_tree_reverse(self):
    """Test scanning up 3 tree levels of ant."""
    sys = test_utils.load_fixture('ant.xml')

    qs, qds = [], []

    def f(parent, link, q, qd):
      if parent is not None:
        link += parent
      qs.append(q)
      qds.append(qd)
      return link

    x = scan.tree(
        sys,
        f,
        'lqd',
        np.arange(sys.num_links()),
        np.arange(sys.q_size()),
        np.arange(sys.qd_size()),
        reverse=True,
    )

    # our scan carry is the sum of index + ancestry indices for each link
    np.testing.assert_array_equal(
        x,
        np.array([
            36,  # 0 + 1 + 2 + 3 + 4 + 5 + 6 + 7 + 8
            3,  # 1 + 2
            2,  # 2
            7,  # 3 + 4
            4,  # 4
            11,  # 5 + 6
            6,  # 6
            15,  # 7 + 8
            8,  # 8
        ]),
    )

    self.assertLen(qs, 3)  # three levels
    np.testing.assert_array_equal(qs[0], np.array([8, 10, 12, 14]))
    np.testing.assert_array_equal(qs[1], np.array([7, 9, 11, 13]))
    np.testing.assert_array_equal(qs[2], np.array([0, 1, 2, 3, 4, 5, 6]))

    self.assertLen(qds, 3)  # three levels
    np.testing.assert_array_equal(qds[0], np.array([7, 9, 11, 13]))
    np.testing.assert_array_equal(qds[1], np.array([6, 8, 10, 12]))
    np.testing.assert_array_equal(qds[2], np.array([0, 1, 2, 3, 4, 5]))

  def test_scan_link_types(self):
    """Test scanning 2 links types of ant."""
    sys = test_utils.load_fixture('ant.xml')

    typs, qs, qds = [], [], []

    def f(typ, link, q, qd):
      typs.append(typ)
      qs.append(q)
      qds.append(qd)
      return link

    x = scan.link_types(
        sys,
        f,
        'lqd',
        'l',
        np.arange(sys.num_links()),
        np.arange(sys.q_size()),
        np.arange(sys.qd_size()),
    )

    self.assertSequenceEqual(typs, ['f', '1'])
    np.testing.assert_array_equal(x, np.array([0, 1, 2, 3, 4, 5, 6, 7, 8]))

    self.assertLen(qs, 2)
    np.testing.assert_array_equal(qs[0], np.arange(0, 7))
    np.testing.assert_array_equal(qs[1], np.arange(7, 15))

    self.assertLen(qds, 2)
    np.testing.assert_array_equal(qds[0], np.arange(0, 6))
    np.testing.assert_array_equal(qds[1], np.arange(6, 14))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Trains a reacher to reach a target.

Based on the OpenAI Gym MuJoCo Reacher environment.
"""

from typing import Tuple

from brax import base
from brax import math
from brax.envs.base import PipelineEnv, State
from brax.io import mjcf
from etils import epath
import jax
from jax import numpy as jp


class Reacher(PipelineEnv):



  # pyformat: disable
  """
  ### Description

  "Reacher" is a two-jointed robot arm. The goal is to move the robot's end
  effector (called *fingertip*) close to a target that is spawned at a random
  position.

  ### Action Space

  The action space is a `Box(-1, 1, (2,), float32)`. An action `(a, b)`
  represents the torques applied at the hinge joints.

  | Num | Action                                                                          | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|---------------------------------------------------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Torque applied at the first hinge (connecting the link to the point of fixture) | -1          | 1           | joint0                         | hinge | torque (N m) |
  | 1   | Torque applied at the second hinge (connecting the two links)                   | -1          | 1           | joint1                         | hinge | torque (N m) |

  ### Observation Space

  Observations consist of:

  - The cosine of the angles of the two arms
  - The sine of the angles of the two arms
  - The coordinates of the target
  - The angular velocities of the arms
  - The vector between the target and the reacher's fingertip (3 dimensional
    with the last element being 0)

  The observation is a `ndarray` with shape `(11,)` where the elements
  correspond to the following:

  | Num | Observation                                                                                    | Min  | Max | Name (in corresponding config) | Joint | Unit                     |
  |-----|------------------------------------------------------------------------------------------------|------|-----|--------------------------------|-------|--------------------------|
  | 0   | cosine of the angle of the first arm                                                           | -Inf | Inf | cos(joint0)                    | hinge | unitless                 |
  | 1   | cosine of the angle of the second arm                                                          | -Inf | Inf | cos(joint1)                    | hinge | unitless                 |
  | 2   | sine of the angle of the first arm                                                             | -Inf | Inf | cos(joint0)                    | hinge | unitless                 |
  | 3   | sine of the angle of the second arm                                                            | -Inf | Inf | cos(joint1)                    | hinge | unitless                 |
  | 4   | x-coordinate of the target                                                                     | -Inf | Inf | target_x                       | slide | position (m)             |
  | 5   | y-coordinate of the target                                                                     | -Inf | Inf | target_y                       | slide | position (m)             |
  | 6   | angular velocity of the first arm                                                              | -Inf | Inf | joint0                         | hinge | angular velocity (rad/s) |
  | 7   | angular velocity of the second arm                                                             | -Inf | Inf | joint1                         | hinge | angular velocity (rad/s) |
  | 8   | x-value of position_fingertip - position_target                                                | -Inf | Inf | NA                             | slide | position (m)             |
  | 9   | y-value of position_fingertip - position_target                                                | -Inf | Inf | NA                             | slide | position (m)             |
  | 10  | z-value of position_fingertip - position_target (0 since reacher is 2d and z is same for both) | -Inf | Inf | NA                             | slide | position (m)             |

  ### Rewards

  The reward consists of two parts:

  - *reward_dist*: This reward is a measure of how far the *fingertip*
    of the reacher (the unattached end) is from the target, with a more negative
    value assigned for when the reacher's *fingertip* is further away from the
    target. It is calculated as the negative vector norm of (position of
    the fingertip - position of target), or *-norm("fingertip" - "target")*.
  - *reward_ctrl*: A negative reward for penalising the walker if it takes
    actions that are too large. It is measured as the negative squared
    Euclidean norm of the action, i.e. as *- sum(action<sup>2</sup>)*.

  Unlike other environments, Reacher does not allow you to specify weights for
  the individual reward terms. However, `info` does contain the keys
  *reward_dist* and *reward_ctrl*. Thus, if you'd like to weight the terms, you
  should create a wrapper that computes the weighted reward from `info`.

  ### Starting State

  All observations start in state (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
  0.0, 0.0) with a noise added for stochasticity. A uniform noise in the range
  [-0.1, 0.1] is added to the positional attributes, while the target position
  is selected uniformly at random in a disk of radius 0.2 around the origin.

  Independent, uniform noise in the range of [-0.005, 0.005] is added to the
  velocities, and the last element ("fingertip" - "target") is calculated at the
  end once everything is set.

  The default setting has a *dt = 0.02*

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches a 1000 timesteps

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder (or by changing
  the path to a modified XML file in another folder)..

  ```
  env = gym.make('Reacher-v2')
  ```

  There is no v3 for Reacher, unlike the robot environments where a v3 and
  beyond take gym.make kwargs such as ctrl_cost_weight, reset_noise_scale etc.

  There is a v4 version that uses the mujoco-bindings

  ```
  env = gym.make('Reacher-v4')
  ```

  And a v5 version that uses Brax:

  ```
  env = gym.make('Reacher-v5')
  ```

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks (not including
    reacher, which has a max_time_steps of 50). Added reward_threshold to
    environments.
  * v0: Initial versions release (1.0.0)
  """
  # pyformat: enable


  def __init__(self, backend='generalized', **kwargs):
    path = epath.resource_path('brax') / 'envs/assets/reacher.xml'
    sys = mjcf.load(path)

    n_frames = 2

    if backend in ['spring', 'positional']:
      sys = sys.tree_replace({'opt.timestep': 0.005})
      sys = sys.replace(
          actuator=sys.actuator.replace(gear=jp.array([25.0, 25.0]))
      )
      n_frames = 4

    kwargs['n_frames'] = kwargs.get('n_frames', n_frames)

    super().__init__(sys=sys, backend=backend, **kwargs)

  def reset(self, rng: jax.Array) -> State:
    rng, rng1, rng2 = jax.random.split(rng, 3)

    q = self.sys.init_q + jax.random.uniform(
        rng1, (self.sys.q_size(),), minval=-0.1, maxval=0.1
    )
    qd = jax.random.uniform(
        rng2, (self.sys.qd_size(),), minval=-0.005, maxval=0.005
    )

    # set the target q, qd
    _, target = self._random_target(rng)
    q = q.at[2:].set(target)
    qd = qd.at[2:].set(0)

    pipeline_state = self.pipeline_init(q, qd)

    obs = self._get_obs(pipeline_state)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'reward_dist': zero,
        'reward_ctrl': zero,
    }
    return State(pipeline_state, obs, reward, done, metrics)

  def step(self, state: State, action: jax.Array) -> State:
    pipeline_state = self.pipeline_step(state.pipeline_state, action)
    obs = self._get_obs(pipeline_state)

    # vector from tip to target is last 3 entries of obs vector
    reward_dist = -math.safe_norm(obs[-3:])
    reward_ctrl = -jp.square(action).sum()
    reward = reward_dist + reward_ctrl

    state.metrics.update(
        reward_dist=reward_dist,
        reward_ctrl=reward_ctrl,
    )

    return state.replace(pipeline_state=pipeline_state, obs=obs, reward=reward)

  def _get_obs(self, pipeline_state: base.State) -> jax.Array:
    """Returns egocentric observation of target and arm body."""
    theta = pipeline_state.q[:2]
    target_pos = pipeline_state.x.pos[2]
    tip_pos = (
        pipeline_state.x.take(1)
        .do(base.Transform.create(pos=jp.array([0.11, 0, 0])))
        .pos
    )
    # tip_vel, instead of pipeline_state.qd[:2], leads to more sensible policies
    # for a randomly initialized policy network
    tip_vel = (
        base.Transform.create(pos=jp.array([0.11, 0, 0]))
        .do(pipeline_state.xd.take(1))
        .vel
    )
    tip_to_target = tip_pos - target_pos

    return jp.concatenate([
        jp.cos(theta),
        jp.sin(theta),
        pipeline_state.q[2:],  # target x, y
        tip_vel[:2],
        tip_to_target,
    ])

  def _random_target(self, rng: jax.Array) -> Tuple[jax.Array, jax.Array]:
    """Returns a target location in a random circle slightly above xy plane."""
    rng, rng1, rng2 = jax.random.split(rng, 3)
    dist = 0.2 * jax.random.uniform(rng1)
    ang = jp.pi * 2.0 * jax.random.uniform(rng2)
    target_x = dist * jp.cos(ang)
    target_y = dist * jp.sin(ang)
    return rng, jp.array([target_x, target_y])



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exports a system config and state as an image."""

import io
from typing import List, Optional, Sequence, Union

import brax
from brax import base
import mujoco
import numpy as np
from PIL import Image


def render_array(
    sys: brax.System,
    trajectory: Union[List[base.State], base.State],
    height: int = 240,
    width: int = 320,
    camera: Optional[str] = None,
) -> Union[Sequence[np.ndarray], np.ndarray]:
  """Returns a sequence of np.ndarray images using the MuJoCo renderer."""
  renderer = mujoco.Renderer(sys.mj_model, height=height, width=width)
  camera = camera or -1

  def get_image(state: base.State):
    d = mujoco.MjData(sys.mj_model)
    d.qpos, d.qvel = state.q, state.qd
    mujoco.mj_forward(sys.mj_model, d)
    renderer.update_scene(d, camera=camera)
    return renderer.render()

  if isinstance(trajectory, list):
    return [get_image(s) for s in trajectory]

  return get_image(trajectory)


def render(
    sys: brax.System,
    trajectory: List[base.State],
    height: int = 240,
    width: int = 320,
    camera: Optional[str] = None,
    fmt: str = 'png',
) -> bytes:
  """Returns an image of a brax System."""
  if not trajectory:
    raise RuntimeError('must have at least one state')

  frames = render_array(sys, trajectory, height, width, camera)
  frames = [Image.fromarray(image) for image in frames]

  f = io.BytesIO()
  if len(frames) == 1:
    frames[0].save(f, format=fmt)
  else:
    frames[0].save(
        f,
        format=fmt,
        append_images=frames[1:],
        save_all=True,
        duration=sys.opt.timestep * 1000,
        loop=0,
    )

  return f.getvalue()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Soft Actor-Critic losses.

See: https://arxiv.org/pdf/1812.05905.pdf
"""
from typing import Any

from brax.training import types
from brax.training.agents.sac import networks as sac_networks
from brax.training.types import Params
from brax.training.types import PRNGKey
import jax
import jax.numpy as jnp

Transition = types.Transition


def make_losses(sac_network: sac_networks.SACNetworks, reward_scaling: float,
                discounting: float, action_size: int):
  """Creates the SAC losses."""

  target_entropy = -0.5 * action_size
  policy_network = sac_network.policy_network
  q_network = sac_network.q_network
  parametric_action_distribution = sac_network.parametric_action_distribution

  def alpha_loss(log_alpha: jnp.ndarray, policy_params: Params,
                 normalizer_params: Any, transitions: Transition,
                 key: PRNGKey) -> jnp.ndarray:
    """Eq 18 from https://arxiv.org/pdf/1812.05905.pdf."""
    dist_params = policy_network.apply(normalizer_params, policy_params,
                                       transitions.observation)
    action = parametric_action_distribution.sample_no_postprocessing(
        dist_params, key)
    log_prob = parametric_action_distribution.log_prob(dist_params, action)
    alpha = jnp.exp(log_alpha)
    alpha_loss = alpha * jax.lax.stop_gradient(-log_prob - target_entropy)
    return jnp.mean(alpha_loss)

  def critic_loss(q_params: Params, policy_params: Params,
                  normalizer_params: Any, target_q_params: Params,
                  alpha: jnp.ndarray, transitions: Transition,
                  key: PRNGKey) -> jnp.ndarray:
    q_old_action = q_network.apply(normalizer_params, q_params,
                                   transitions.observation, transitions.action)
    next_dist_params = policy_network.apply(normalizer_params, policy_params,
                                            transitions.next_observation)
    next_action = parametric_action_distribution.sample_no_postprocessing(
        next_dist_params, key)
    next_log_prob = parametric_action_distribution.log_prob(
        next_dist_params, next_action)
    next_action = parametric_action_distribution.postprocess(next_action)
    next_q = q_network.apply(normalizer_params, target_q_params,
                             transitions.next_observation, next_action)
    next_v = jnp.min(next_q, axis=-1) - alpha * next_log_prob
    target_q = jax.lax.stop_gradient(transitions.reward * reward_scaling +
                                     transitions.discount * discounting *
                                     next_v)
    q_error = q_old_action - jnp.expand_dims(target_q, -1)

    # Better bootstrapping for truncated episodes.
    truncation = transitions.extras['state_extras']['truncation']
    q_error *= jnp.expand_dims(1 - truncation, -1)

    q_loss = 0.5 * jnp.mean(jnp.square(q_error))
    return q_loss

  def actor_loss(policy_params: Params, normalizer_params: Any,
                 q_params: Params, alpha: jnp.ndarray, transitions: Transition,
                 key: PRNGKey) -> jnp.ndarray:
    dist_params = policy_network.apply(normalizer_params, policy_params,
                                       transitions.observation)
    action = parametric_action_distribution.sample_no_postprocessing(
        dist_params, key)
    log_prob = parametric_action_distribution.log_prob(dist_params, action)
    action = parametric_action_distribution.postprocess(action)
    q_action = q_network.apply(normalizer_params, q_params,
                               transitions.observation, action)
    min_q = jnp.min(q_action, axis=-1)
    actor_loss = alpha * log_prob - min_q
    return jnp.mean(actor_loss)

  return alpha_loss, critic_loss, actor_loss



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""setup.py for Brax.

Install for development:

  pip intall -e .
"""

from setuptools import find_packages
from setuptools import setup

setup(
    name="brax",
    version="0.10.5",
    description="A differentiable physics engine written in JAX.",
    author="Brax Authors",
    author_email="no-reply@google.com",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="http://github.com/google/brax",
    license="Apache 2.0",
    packages=find_packages(),
    include_package_data=True,
    scripts=["bin/learn"],
    install_requires=[
        "absl-py",
        "dataclasses; python_version < '3.7'",
        # TODO: remove dm_env after dropping legacy v1 code
        "dm_env",
        "etils",
        "flask",
        "flask_cors",
        "flax",
        # TODO: remove grpcio and gym after dropping legacy v1 code
        "grpcio",
        "gym",
        "jax>=0.4.6",
        "jaxlib>=0.4.6",
        "jaxopt",
        "jinja2",
        "ml_collections",
        "mujoco",
        "mujoco-mjx",
        "numpy",
        "optax",
        "orbax-checkpoint",
        # TODO: remove pytinyrenderer after dropping legacy v1 code
        "Pillow",
        "pytinyrenderer",
        "scipy",
        "tensorboardX",
        "trimesh",
        "typing-extensions",
    ],
    extras_require={
        "develop": ["pytest", "transforms3d"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="JAX reinforcement learning rigidbody physics",
)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for adversarial IRL and state marginal matching RL.

See:
  "A Divergence Minimization Perspective on Imitation Learning"
  https://arxiv.org/abs/1911.02256
"""

import functools
import itertools
from typing import Any, Callable, List, Tuple, Optional, Dict
from brax.v1.envs.env import Env
from brax.v1.envs.env import State
from brax.v1.experimental import normalization
from brax.v1.experimental.braxlines.common import dist_utils
from brax.v1.experimental.composer import composer
from brax.v1.experimental.composer import observers
from brax.training import networks
from brax.training.ppo import StepData
import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp

tfd = tfp.distributions

DISC_PARAM_NAME = "irl_disc_params"


class IRLDiscriminator(object):
  """Discriminator for target data versus on-policy data."""

  def __init__(
      self,
      env: Env,
      input_size: int,
      reward_type: str = "gail",
      arch: Tuple[int] = (32, 32),
      obs_indices: Optional[List[Any]] = None,
      act_indices: Optional[List[int]] = None,
      obs_scale: Optional[List[float]] = None,
      include_action: bool = False,
      logits_clip_range: float = 10.0,
      nonnegative_reward: bool = True,
      param_name: str = DISC_PARAM_NAME,
      target_data: jnp.ndarray = None,
      target_dist_fn=None,
      balance_data: bool = True,
      normalize_obs: bool = False,
      spectral_norm: bool = False,
      gradient_penalty_weight: float = 0.,
  ):
    assert obs_scale is not None
    self.env_obs_size = env.observation_size
    self.arch = arch
    self.input_size = input_size
    self.dist_fn = lambda x: dist_utils.clipped_bernoulli(
        logits=x, clip_range=logits_clip_range)
    self.obs_indices, self.obs_labels = observers.index_preprocess(
        obs_indices, env)
    self.indexed_obs_size = len(self.obs_indices)
    self.obs_scale = jnp.array(obs_scale or 1.) * jnp.ones(
        self.indexed_obs_size)
    if self.indexed_obs_size == 1:
      self.obs_labels_2d = (self.obs_labels[0], "None")
      self.obs_scale_2d = jnp.concatenate([self.obs_scale, self.obs_scale])
    else:
      self.obs_labels_2d = self.obs_labels[:2]
      self.obs_scale_2d = self.obs_scale[:2]
    self.act_indices = act_indices
    self.include_action = include_action
    self.reward_type = reward_type
    self.param_name = param_name
    self.model = None
    self.initialized = False
    self.target_data = target_data
    self.target_dist_fn = target_dist_fn
    if self.reward_type == "mle":
      assert self.target_dist_fn
    self.logits_clip_range = logits_clip_range
    self.nonnegative_reward = nonnegative_reward
    if self.nonnegative_reward:
      if self.reward_type == "mle":
        # approximately find the global min
        edges_mesh = itertools.product(*([(-1, 0, 1)] * self.indexed_obs_size))
        edges_mesh = jnp.stack([jnp.array(x) for x in edges_mesh])
        x_edges = edges_mesh * self.obs_scale[None]
        x_target_dist = target_dist_fn()
        x_samples = x_target_dist.sample(
            seed=jax.random.PRNGKey(0), sample_shape=(1000,))
        x = jnp.concatenate([x_edges, x_samples], axis=0)
        r = target_dist_fn().log_prob(x)
        self.target_dist_log_offset = -r.min()
      else:
        assert self.logits_clip_range
    if gradient_penalty_weight:
      balance_data = True
    self.balance_data = balance_data
    self.normalize_obs = normalize_obs
    self.normalize_fn = normalization.make_data_and_apply_fn(
        [self.env_obs_size], self.normalize_obs)[1]
    self.spectral_norm = spectral_norm
    self.gradient_penalty_weight = gradient_penalty_weight

  def set_target_data(self, target_data: jnp.ndarray):
    self.target_data = target_data

  def init_model(self, rng: jnp.ndarray = None):
    """Initialize neural network modules."""
    model = networks.make_model(
        self.arch + (1,), self.input_size, spectral_norm=self.spectral_norm)
    self.model = model
    if self.spectral_norm:
      rng1, rng2, rng3 = jax.random.split(rng, 3)
      model_params = model.init(rng1, rng2)
      self.fn = lambda params, x: model.apply(
          params, x, rngs={"sing_vec": rng3}, mutable=["sing_vec"])[0]
    else:
      model_params = model.init(rng)
      self.fn = model.apply

    def mean_fn(params, x):
      return jnp.mean(self.fn(params, x))

    self.dmean_fn_dx = jax.grad(mean_fn, argnums=1)

    self.model = model
    self.initialized = True
    return {self.param_name: model_params}

  def dist(self,
           data: jnp.ndarray,
           params: Dict[str, Dict[str, jnp.ndarray]] = None):
    assert self.initialized, "init_model() must be called"
    param = params[self.param_name]
    return self.dist_fn(self.fn(param, data))

  def ll(self,
         data: jnp.ndarray,
         labels: jnp.ndarray,
         params: Dict[str, Dict[str, jnp.ndarray]] = None) -> jnp.ndarray:
    dist = self.dist(data, params=params)
    ll = dist.log_prob(labels)
    ll = jnp.sum(ll, axis=-1)
    return ll

  def irl_reward(
      self,
      data: jnp.ndarray,
      params: Dict[str, Dict[str, jnp.ndarray]] = None) -> jnp.ndarray:
    """Compute IRL reward."""
    dist = self.dist(data, params=params)
    if self.reward_type == "gail":
      r = -dist.log_prob(jnp.zeros_like(dist.logits))
      r = jnp.sum(r, axis=-1)
    elif self.reward_type == "gail2":
      # https://arxiv.org/abs/2106.00672
      r = dist.log_prob(jnp.ones_like(dist.logits))
      if self.nonnegative_reward:
        r += self.logits_clip_range
      r = jnp.sum(r, axis=-1)
    elif self.reward_type == "airl":
      r = dist.logits
      if self.nonnegative_reward:
        r += self.logits_clip_range
      r = jnp.sum(r, axis=-1)
    elif self.reward_type == "fairl":
      r = dist.logits
      r = jnp.exp(r) * -r
      if self.nonnegative_reward:
        r += self.logits_clip_range
      r = jnp.sum(r, axis=-1)
    elif self.reward_type == "mle":  # for debugging
      assert not self.normalize_obs
      target_dist = self.target_dist_fn()
      r = target_dist.log_prob(data)
      if self.nonnegative_reward:
        r += self.target_dist_log_offset
    else:
      raise NotImplementedError(self.reward_type)
    return r

  def obs_act2data(self, obs: jnp.ndarray, act: jnp.ndarray):
    """Convert obs and actions into data."""
    assert obs.shape[:-1] == act.shape[:-1], f"obs={obs.shape}, act={act.shape}"
    data = obs
    data = self.index_obs(data)
    if self.include_action:
      if self.act_indices:
        act = act[..., self.act_indices]
      data = jnp.concatenate([data, act], axis=-1)
    return data

  def index_obs(self, obs: jnp.ndarray):
    if self.obs_indices:
      return obs[..., self.obs_indices]
    else:
      return obs

  def disc_loss_fn(self, data: StepData, udata: StepData, rng: jnp.ndarray,
                   params: Dict[str, Dict[str, jnp.ndarray]]):
    return disc_loss_fn(
        data,
        udata,
        rng,
        params,
        disc=self,
        normalize_obs=self.normalize_obs,
        target_data=self.target_data,
        balance_data=self.balance_data,
        gradient_penalty_weight=self.gradient_penalty_weight,
    )


class IRLWrapper(Env):
  """A wrapper that adds an IRL reward to a Brax Env."""

  def __init__(
      self,
      environment: Env,
      disc: IRLDiscriminator,
      env_reward_multiplier: float = 0.0,
  ):
    self._environment = environment
    self.action_repeat = self._environment.action_repeat
    if hasattr(self._environment, "batch_size"):
      self.batch_size = self._environment.batch_size
    else:
      self.batch_size = None
    self.sys = self._environment.sys
    self.disc = disc
    self.env_reward_multiplier = env_reward_multiplier

  def reset(self, rng: jnp.ndarray) -> State:
    """Resets the environment to an initial state."""
    state = self._environment.reset(rng)
    return state.replace(reward=jnp.zeros_like(state.reward))

  def step(self,
           state: State,
           action: jnp.ndarray,
           normalizer_params: Dict[str, jnp.ndarray] = None,
           extra_params: Dict[str, Dict[str, jnp.ndarray]] = None) -> State:
    """Run one timestep of the environment's dynamics."""
    obs = self.disc.normalize_fn(normalizer_params, state.obs)
    new_reward = disc_reward_fn(
        obs, action, params=extra_params, disc=self.disc)
    state = self._environment.step(state, action)
    return state.replace(reward=new_reward +
                         self.env_reward_multiplier * state.reward)


def disc_reward_fn(
    obs: jnp.ndarray,
    actions: jnp.ndarray,
    params: Dict[str, Dict[str, jnp.ndarray]],
    disc: IRLDiscriminator,
) -> jnp.ndarray:
  data = disc.obs_act2data(obs, actions)
  new_reward = disc.irl_reward(data, params)
  return jax.lax.stop_gradient(new_reward)


def disc_loss_fn(
    data: StepData,
    udata: StepData,
    rng: jnp.ndarray,
    params: Dict[str, Dict[str, jnp.ndarray]],
    disc: IRLDiscriminator,
    target_data: jnp.ndarray,
    balance_data: bool = True,
    normalize_obs: bool = False,
    gradient_penalty_weight: float = 0.,
):
  """Discriminator loss function."""
  d = data if normalize_obs else udata
  target_d = target_data  # TODO: add normalize option for target_data
  d = disc.obs_act2data(d.obs[:d.actions.shape[0]], d.actions)
  d = jnp.reshape(d, [-1, d.shape[-1]])
  target_d = jnp.reshape(target_d, [-1, target_d.shape[-1]])
  if balance_data and d.shape[0] != target_d.shape[0]:
    rng, loss_key = jax.random.split(rng)
    if d.shape[0] > target_d.shape[0]:
      indices = jnp.arange(0, d.shape[0])
      indices = jax.random.permutation(loss_key, indices)
      d = d[indices[:target_d.shape[0]]]
    else:
      indices = jax.random.permutation(loss_key, target_d.shape[0])
      target_d = target_d[indices[:d.shape[0]]]
  disc_loss = -jnp.mean(
      disc.ll(d, jnp.zeros(d.shape[:-1] + (1,)), params=params))
  disc_loss += -jnp.mean(
      disc.ll(target_d, jnp.ones(target_d.shape[:-1] + (1,)), params=params))

  if gradient_penalty_weight > 0.:
    assert (
        d.shape == target_d.shape
    ), f"d shape {d.shape} does not match target_d shape {target_d.shape}!"
    rng, sub_key = jax.random.split(rng)
    w_shape = [d.shape[0], 1]
    w = jax.random.uniform(sub_key, shape=w_shape, minval=0., maxval=1.)
    interp_d = w * d + (1. - w) * target_d

    p = params[disc.param_name]
    grad = disc.dmean_fn_dx(p, interp_d)
    if len(grad.shape) > 2:
      grad = jnp.reshape(grad, [grad.shape[0], -1])
    grad_norm = jnp.sum(jnp.linalg.norm(grad, axis=-1))
    disc_loss = disc_loss + gradient_penalty_weight * grad_norm

  return disc_loss, rng


def create(env_name: str, wrapper_params: Dict[str, Any], **kwargs) -> Env:
  """Creates an Env with a specified brax system."""
  env = composer.create(env_name, **kwargs)
  return IRLWrapper(env, **wrapper_params)


def create_fn(env_name: str, wrapper_params: Dict[str, Any],
              **kwargs) -> Callable[..., Env]:
  """Returns a function that when called, creates an Env."""
  return functools.partial(
      create, env_name=env_name, wrapper_params=wrapper_params, **kwargs)


def make_2d(data):
  """Make data to be 2D in last dimension."""
  if data.shape[-1] == 1:
    return jnp.concatenate([data, jnp.zeros_like(data)], axis=-1)
  else:
    return data[..., :2]


def get_multimode_dist(num_modes: int = 1,
                       scale: float = 1.0,
                       indexed_obs_dim: int = 1):
  """Get a multimodal distribution of Gaussians."""
  if indexed_obs_dim == 1:
    return get_multimode_1d_dist(num_modes=num_modes, scale=scale)
  elif indexed_obs_dim == 2:
    return get_multimode_2d_dist(num_modes=num_modes, scale=scale)
  else:
    raise NotImplementedError(indexed_obs_dim)


def get_multimode_1d_dist(num_modes: int = 1, scale: float = 1.0):
  """Get a multimodal distribution of Gaussians."""
  scale = jnp.ones(1) * scale
  loc = jnp.linspace(-scale, scale, num_modes + 2)
  loc = loc[1:-1]
  scale = jnp.ones((num_modes, 1)) * scale[None] / (num_modes + 1) / 5.
  return tfd.MixtureSameFamily(
      mixture_distribution=tfd.Categorical(
          probs=jnp.ones((num_modes,)) / num_modes),
      components_distribution=tfd.MultivariateNormalDiag(
          loc=loc, scale_diag=scale))


def get_multimode_2d_dist(num_modes: int = 1, scale: float = 1.0):
  """Get a multimodal distribution of Gaussians."""
  angles = jnp.linspace(0, jnp.pi * 2, num_modes + 1)
  angles = angles[:-1]
  x, y = jnp.cos(angles) * scale / 2., jnp.sin(angles) * scale / 2.
  loc = jnp.array([x, y]).T
  scale = jnp.ones((num_modes, 2)) * scale / 10.
  return tfd.MixtureSameFamily(
      mixture_distribution=tfd.Categorical(
          probs=jnp.ones((num_modes,)) / num_modes),
      components_distribution=tfd.MultivariateNormalDiag(
          loc=loc, scale_diag=scale))



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Some example environments to help get started quickly with brax."""

import functools
from typing import Callable, Optional, Type, Union, overload

import brax.v1  # pylint:disable=unused-import

from brax.v1.envs import acrobot
from brax.v1.envs import ant
from brax.v1.envs import fast
from brax.v1.envs import fetch
from brax.v1.envs import grasp
from brax.v1.envs import half_cheetah
from brax.v1.envs import hopper
from brax.v1.envs import humanoid
from brax.v1.envs import humanoid_standup
from brax.v1.envs import inverted_double_pendulum
from brax.v1.envs import inverted_pendulum
from brax.v1.envs import pusher
from brax.v1.envs import reacher
from brax.v1.envs import reacherangle
from brax.v1.envs import swimmer
from brax.v1.envs import ur5e
from brax.v1.envs import walker2d
from brax.v1.envs import wrappers
from brax.v1.envs.env import Env, State, Wrapper
import gym

_envs = {
    'acrobot': acrobot.Acrobot,
    'ant': functools.partial(ant.Ant, use_contact_forces=True),
    'fast': fast.Fast,
    'fetch': fetch.Fetch,
    'grasp': grasp.Grasp,
    'halfcheetah': half_cheetah.Halfcheetah,
    'hopper': hopper.Hopper,
    'humanoid': humanoid.Humanoid,
    'humanoidstandup': humanoid_standup.HumanoidStandup,
    'inverted_pendulum': inverted_pendulum.InvertedPendulum,
    'inverted_double_pendulum': inverted_double_pendulum.InvertedDoublePendulum,
    'pusher': pusher.Pusher,
    'reacher': reacher.Reacher,
    'reacherangle': reacherangle.ReacherAngle,
    'swimmer': swimmer.Swimmer,
    'ur5e': ur5e.Ur5e,
    'walker2d': walker2d.Walker2d,
}



def get_environment(env_name, **kwargs) -> Env:
  return _envs[env_name](**kwargs)


def register_environment(env_name: str, env_class: Type[Env]):
  _envs[env_name] = env_class


def create(env_name: str,
           episode_length: int = 1000,
           action_repeat: int = 1,
           auto_reset: bool = True,
           batch_size: Optional[int] = None,
           eval_metrics: bool = False,
           **kwargs) -> Env:
  """Creates an Env with a specified brax system."""
  env = _envs[env_name](**kwargs)
  if episode_length is not None:
    env = wrappers.EpisodeWrapper(env, episode_length, action_repeat)
  if batch_size:
    env = wrappers.VectorWrapper(env, batch_size)
  if auto_reset:
    env = wrappers.AutoResetWrapper(env)
  if eval_metrics:
    env = wrappers.EvalWrapper(env)

  return env  # type: ignore


def create_fn(env_name: str, **kwargs) -> Callable[..., Env]:
  """Returns a function that when called, creates an Env."""
  return functools.partial(create, env_name, **kwargs)


@overload
def create_gym_env(env_name: str,
                   batch_size: None = None,
                   seed: int = 0,
                   backend: Optional[str] = None,
                   **kwargs) -> gym.Env:
  ...


@overload
def create_gym_env(env_name: str,
                   batch_size: int,
                   seed: int = 0,
                   backend: Optional[str] = None,
                   **kwargs) -> gym.vector.VectorEnv:
  ...


def create_gym_env(env_name: str,
                   batch_size: Optional[int] = None,
                   seed: int = 0,
                   backend: Optional[str] = None,
                   **kwargs) -> Union[gym.Env, gym.vector.VectorEnv]:
  """Creates a `gym.Env` or `gym.vector.VectorEnv` from a Brax environment."""
  environment = create(env_name=env_name, batch_size=batch_size, **kwargs)
  if batch_size is None:
    return wrappers.GymWrapper(environment, seed=seed, backend=backend)
  if batch_size <= 0:
    raise ValueError(
        '`batch_size` should either be None or a positive integer.')
  return wrappers.VectorGymWrapper(environment, seed=seed, backend=backend)
